# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2013 Oct 30
from __future__ import print_function
import os
import matplotlib

print('matplotlib version: ', matplotlib.__version__)

import gpi
from gpi import QtCore, QtGui, QtWidgets

import numpy as np
from matplotlib.figure import Figure
#from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_qt5agg import (
    FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
from matplotlib.backends.backend_qt5 import SubplotToolQt

class MainWin_close(QtWidgets.QMainWindow):
    window_closed = gpi.Signal()
    def __init__(self):
        super().__init__()
        self._isActive = True

    def closeEvent(self, event):
        super().closeEvent(event)
        self.window_closed.emit()
        self._isActive = False

    def isActive(self):
        return self._isActive

class NavbarTools(NavigationToolbar):
    # list of toolitems to add to the toolbar, format is:
    # (
    #   text, # the text of the button (often not visible to users)
    #   tooltip_text, # the tooltip shown on hover (where possible)
    #   image_file, # name of the image for the button (without the extension)
    #   name_of_method, # name of the method in NavigationToolbar2 to call
    # )
    toolitems = (
        ('Home', 'Reset original view', 'home', 'home'),
        ('Back', 'Back to  previous view', 'back', 'back'),
        ('Forward', 'Forward to next view', 'forward', 'forward'),
        (None, None, None, None),
        ('Pan', 'Pan axes with left mouse, zoom with right', 'move', 'pan'),
        ('Zoom', 'Zoom to rectangle', 'zoom_to_rect', 'zoom'),
        (None, None, None, None),
        #('Subplots', 'Configure subplots', 'subplots', 'configure_subplots'),
        ('Save', 'Save the figure', 'filesave', 'save_figure'),
      )

    def __init__(self, canvas, parent):
        super().__init__(canvas, parent)

    def _init_toolbar(self):
        self.basedir = os.path.join(matplotlib.rcParams[ 'datapath' ],'images')

        for text, tooltip_text, image_file, callback in self.toolitems:
            if text is None:
                self.addSeparator()
            else:
                a = self.addAction(self._icon(image_file + '.png'),
                                         text, getattr(self, callback))
                self._actions[callback] = a
                if callback in ['zoom', 'pan']:
                    a.setCheckable(True)
                if tooltip_text is not None:
                    a.setToolTip(tooltip_text)

        # Add the x,y location widget at the right side of the toolbar
        # The stretch factor is 1 which means any resizing of the toolbar
        # will resize this label instead of the buttons.
        if self.coordinates:
            self.locLabel = QtWidgets.QLabel( "", self )
            self.locLabel.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignTop )
            self.locLabel.setSizePolicy(
                QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Ignored))
            labelAction = self.addWidget(self.locLabel)
            labelAction.setVisible(True)

###############################################################################
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 Pierre Raybau
# Licensed under the terms of the MIT License
# see the mpl licenses directory for a copy of the license

"""Module that provides a GUI-based editor for matplotlib's figure options"""

#from __future__ import print_function
import os.path as osp

try:
    import matplotlib.backends.qt_editor.formlayout as formlayout
except:
    formlayout = None
    print("formlayout can't be found, line options will be disabled")

from matplotlib import markers

def get_icon(name):
    basedir = osp.join(matplotlib.rcParams['datapath'], 'images')
    return QtGui.QIcon(osp.join(basedir, name))

LINESTYLES = {
              '-': 'Solid',
              '--': 'Dashed',
              '-.': 'DashDot',
              ':': 'Dotted',
              'steps': 'Steps',
              'none': 'None',
              }

MARKERS = markers.MarkerStyle.markers

COLORS = {'b': '#0000ff', 'g': '#00ff00', 'r': '#ff0000', 'c': '#ff00ff',
          'm': '#ff00ff', 'y': '#ffff00', 'k': '#000000', 'w': '#ffffff'}

def col2hex(color):
    """Convert matplotlib color to hex"""
    return COLORS.get(color, color)

def figure_edit(axes, parent=None):
    """Edit matplotlib figure options"""
    sep = (None, None) # separator

    has_curve = len(axes.get_lines()) > 0

    # Get / General
    #xmin, xmax = axes.get_xlim()
    #ymin, ymax = axes.get_ylim()
    #general = [('Title', axes.get_title()),
    #           sep,
    #           (None, "<b>X-Axis</b>"),
    #           ('Min', xmin), ('Max', xmax),
    #           ('Label', axes.get_xlabel()),
    #           ('Scale', [axes.get_xscale(), 'linear', 'log']),
    #           sep,
    #           (None, "<b>Y-Axis</b>"),
    #           ('Min', ymin), ('Max', ymax),
    #           ('Label', axes.get_ylabel()),
    #           ('Scale', [axes.get_yscale(), 'linear', 'log'])
    #           ]

    if not has_curve:
        return

    if has_curve:
        # Get / Curves
        linedict = {}
        for line in axes.get_lines():
            label = line.get_label()
            if label == '_nolegend_':
                continue
            linedict[label] = line
        curves = []
        linestyles = LINESTYLES.items()
        markers = MARKERS.items()
        curvelabels = sorted(linedict.keys())
        for label in curvelabels:
            line = linedict[label]
            curvedata = [
                         ('Label', label),
                         sep,
                         (None, '<b>Line</b>'),
                         ('Style', [line.get_linestyle()] + list(linestyles)),
                         ('Width', line.get_linewidth()),
                         ('Color', col2hex(line.get_color())),
                         sep,
                         (None, '<b>Marker</b>'),
                         ('Style', [line.get_marker()] + list(markers)),
                         ('Size', line.get_markersize()),
                         ('Facecolor', col2hex(line.get_markerfacecolor())),
                         ('Edgecolor', col2hex(line.get_markeredgecolor())),
                         ]
            curves.append([curvedata, label, ""])

    #datalist = [(general, "Axes", "")]
    #if has_curve:
    #datalist.append((curves, "Curves", ""))
    datalist = [(curves, "Curves", "")]

    def apply_callback(data):
        """This function will be called to apply changes"""
        #if has_curve:
        #    general, curves = data
        #else:
        #    general, = data

        curves = data[0]

        # Set / General
        #title, xmin, xmax, xlabel, xscale, ymin, ymax, ylabel, yscale = general
        #axes.set_xscale(xscale)
        #axes.set_yscale(yscale)
        #axes.set_title(title)
        #axes.set_xlim(xmin, xmax)
        #axes.set_xlabel(xlabel)
        #axes.set_ylim(ymin, ymax)
        #axes.set_ylabel(ylabel)

        if has_curve:
            # Set / Curves
            for index, curve in enumerate(curves):
                line = linedict[curvelabels[index]]
                label, linestyle, linewidth, color, \
                    marker, markersize, markerfacecolor, markeredgecolor = curve
                line.set_label(label)
                line.set_linestyle(linestyle)
                line.set_linewidth(linewidth)
                line.set_color(color)
                if marker != 'none':
                    line.set_marker(marker)
                    line.set_markersize(markersize)
                    line.set_markerfacecolor(markerfacecolor)
                    line.set_markeredgecolor(markeredgecolor)

        # Redraw
        figure = axes.get_figure()
        figure.canvas.draw()

    if formlayout is not None:
        data = formlayout.fedit(datalist, title="Figure options", parent=parent, icon=get_icon('qt4_editor_options.svg'), apply=apply_callback)

        if data is not None:
            apply_callback(data)
###############################################################################


class MatplotDisplay(gpi.GenericWidgetGroup):

    """Embeds the matplotlib figure window.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        # gpi interface
        self._collapsables = []
        self._subplotSettings = {}
        #self._subplotPosition = {'right': 0.9, 'bottom': 0.12, 'top': 0.9, 'wspace': 0.2, 'hspace': 0.2, 'left': 0.125}
        self._subplotPosition = {'right': 0.913, 'bottom': 0.119, 'top': 0.912, 'wspace': 0.2, 'hspace': 0.2, 'left': 0.111}
        #self._subplot_keepers = ['yscale', 'xscale'] # linear, log
        self._subplot_keepers = []
        self._lineSettings = []
        self._line_keepers = ['linewidth', 'linestyle', 'label', 'marker', 'markeredgecolor', 'markerfacecolor', 'markersize', 'color', 'alpha']

        # since drawing is slow, don't do it as often, use the timer as a
        # debouncer
        self._on_draw_cnt = 0
        self._updatetimer = QtCore.QTimer()
        self._updatetimer.setSingleShot(True)
        self._updatetimer.timeout.connect(self._on_draw)
        self._updatetimer.setInterval(10)

        # plot specific UI side panel
        #  -sets options for plot window so this needs to be run first
        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)  # no spaces around this item
        vbox.setSpacing(0)

        # AUTOSCALE
        self._autoscale_btn = gpi.widgets.BasicPushButton(self)
        self._autoscale_btn.set_toggle(True)
        self._autoscale_btn.set_button_title('autoscale')
        self._autoscale_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._autoscale_btn)

        # GRID
        self._grid_btn = gpi.widgets.BasicPushButton(self)
        self._grid_btn.set_toggle(True)
        self._grid_btn.set_button_title('grid')
        self._grid_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._grid_btn)

        # X/Y LIMITS
        lims = QtWidgets.QGridLayout()
        self._xl = gpi.widgets.BasicDoubleSpinBox(self)
        self._xh = gpi.widgets.BasicDoubleSpinBox(self)
        self._yl = gpi.widgets.BasicDoubleSpinBox(self)
        self._yh = gpi.widgets.BasicDoubleSpinBox(self)
        self._xl.valueChanged.connect(self.on_draw)
        self._xh.valueChanged.connect(self.on_draw)
        self._yl.valueChanged.connect(self.on_draw)
        self._yh.valueChanged.connect(self.on_draw)
        self._xl.set_immediate(True)
        self._xh.set_immediate(True)
        self._yl.set_immediate(True)
        self._yh.set_immediate(True)
        self._xl.set_label('max')
        self._xh.set_label('min')
        self._xl.set_decimals(7)
        self._xh.set_decimals(7)
        self._yl.set_decimals(7)
        self._yh.set_decimals(7)
        self._xlab = QtWidgets.QLabel('x limits')
        self._ylab = QtWidgets.QLabel('y limits')
        #self._maxlab = QtWidgets.QLabel('max')
        #self._minlab = QtWidgets.QLabel('min')
        #lims.addWidget(self._maxlab,1,0,1,1)
        #lims.addWidget(self._minlab,2,0,1,1)
        lims.addWidget(self._xlab,0,1,1,1,alignment=QtCore.Qt.AlignHCenter)
        lims.addWidget(self._xh,1,1,1,1,alignment=QtCore.Qt.AlignHCenter)
        lims.addWidget(self._xl,2,1,1,1,alignment=QtCore.Qt.AlignHCenter)
        lims.addWidget(self._ylab,0,2,1,1,alignment=QtCore.Qt.AlignHCenter)
        lims.addWidget(self._yh,1,2,1,1,alignment=QtCore.Qt.AlignHCenter)
        lims.addWidget(self._yl,2,2,1,1,alignment=QtCore.Qt.AlignHCenter)
        self._collapsables.append(self._xlab)
        self._collapsables.append(self._ylab)
        self._collapsables.append(self._xl)
        self._collapsables.append(self._xh)
        self._collapsables.append(self._yl)
        self._collapsables.append(self._yh)
        #self._collapsables.append(self._minlab)
        #self._collapsables.append(self._maxlab)

        # TICK MARKS
        ticks = QtWidgets.QGridLayout()
        self._x_numticks = gpi.widgets.BasicSpinBox(self)
        self._x_numticks.valueChanged.connect(self.on_draw)
        self._y_numticks = gpi.widgets.BasicSpinBox(self)
        self._y_numticks.valueChanged.connect(self.on_draw)
        self._x_ticks = QtWidgets.QLineEdit()
        self._y_ticks = QtWidgets.QLineEdit()
        self._x_ticks.textChanged.connect(lambda txt: self.check_validticks(self._x_ticks))
        self._y_ticks.textChanged.connect(lambda txt: self.check_validticks(self._y_ticks))
        self._x_ticks.setPlaceholderText('comma separated list of x labels')
        self._y_ticks.setPlaceholderText('comma separated list of y labels')
        self._x_ticks.returnPressed.connect(self.on_draw)
        self._y_ticks.returnPressed.connect(self.on_draw)
        self._x_numticks.set_immediate(True)
        self._y_numticks.set_immediate(True)
        self._x_numticks.set_min(2)
        self._y_numticks.set_min(2)
        self._x_numticks.set_max(100)
        self._y_numticks.set_max(100)
        self._x_numticks.set_val(5)
        self._y_numticks.set_val(5)
        self._x_numticks.set_label('x ticks')
        self._y_numticks.set_label('y ticks')
        ticks.addWidget(self._x_numticks, 0,0,1,1)
        ticks.addWidget(self._y_numticks, 1,0,1,1)
        ticks.addWidget(self._x_ticks, 0,1,1,1)
        ticks.addWidget(self._y_ticks, 1,1,1,1)
        self._collapsables.append(self._x_numticks)
        self._collapsables.append(self._y_numticks)
        self._collapsables.append(self._x_ticks)
        self._collapsables.append(self._y_ticks)

        # TITLE, XLABEL, YLABEL
        plotlabels = QtWidgets.QHBoxLayout()
        self._plot_title = QtWidgets.QLineEdit()
        self._plot_xlab = QtWidgets.QLineEdit()
        self._plot_ylab = QtWidgets.QLineEdit()
        self._plot_title.setPlaceholderText('title')
        self._plot_xlab.setPlaceholderText('x label')
        self._plot_ylab.setPlaceholderText('y label')
        self._plot_title.returnPressed.connect(self.on_draw)
        self._plot_xlab.returnPressed.connect(self.on_draw)
        self._plot_ylab.returnPressed.connect(self.on_draw)
        plotlabels.addWidget(self._plot_title)
        plotlabels.addWidget(self._plot_xlab)
        plotlabels.addWidget(self._plot_ylab)
        self._collapsables.append(self._plot_title)
        self._collapsables.append(self._plot_xlab)
        self._collapsables.append(self._plot_ylab)

        # XSCALE, YSCALE
        self._xscale_btn = gpi.widgets.BasicPushButton(self)
        self._xscale_btn.set_toggle(True)
        self._xscale_btn.set_button_title('log(x)')
        self._xscale_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._xscale_btn)
        self._yscale_btn = gpi.widgets.BasicPushButton(self)
        self._yscale_btn.set_toggle(True)
        self._yscale_btn.set_button_title('log(y)')
        self._yscale_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._yscale_btn)

        scale_options_layout = QtWidgets.QHBoxLayout()
        scale_options_layout.addWidget(self._xscale_btn)
        scale_options_layout.addWidget(self._yscale_btn)

        # LEGEND
        self._legend_btn = gpi.widgets.BasicPushButton(self)
        self._legend_btn.set_toggle(True)
        self._legend_btn.set_button_title('legend')
        self._legend_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._legend_btn)

        # HOLD
        self._hold_btn = gpi.widgets.BasicPushButton(self)
        self._hold_btn.set_toggle(True)
        self._hold_btn.set_button_title('hold')
        #self._hold_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._hold_btn)

        # MOVE AXES TO ORIGIN
        # self._origin_axes_btn = gpi.widgets.BasicPushButton(self)
        # self._origin_axes_btn.set_toggle(True)
        # self._origin_axes_btn.set_button_title("axes at (0,0)")
        # self._collapsables.append(self._origin_axes_btn)

        # RESET
        self._reset_btn = gpi.widgets.BasicPushButton(self)
        self._reset_btn.set_toggle(False)
        self._reset_btn.set_button_title('reset')
        self._reset_btn.valueChanged.connect(self._init_parms_)
        self._collapsables.append(self._reset_btn)

        # X=0, Y=0
        self._xeq0_btn = gpi.widgets.BasicPushButton(self)
        self._xeq0_btn.set_toggle(True)
        self._xeq0_btn.set_button_title('x=0')
        self._xeq0_btn.set_val(True)
        self._xeq0_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._xeq0_btn)
        self._yeq0_btn = gpi.widgets.BasicPushButton(self)
        self._yeq0_btn.set_toggle(True)
        self._yeq0_btn.set_button_title('y=0')
        self._yeq0_btn.set_val(True)
        self._yeq0_btn.valueChanged.connect(self.on_draw)
        self._collapsables.append(self._yeq0_btn)

        # LINE OPTIONS
        self._lino_btn = gpi.widgets.BasicPushButton(self)
        self._lino_btn.set_toggle(False)
        self._lino_btn.set_button_title('line options')
        self._lino_btn.valueChanged.connect(self.lineOptionsDialog)
        self._collapsables.append(self._lino_btn)

        # SUBPLOT SPACING OPTIONS
        self._subplotso_btn = gpi.widgets.BasicPushButton(self)
        self._subplotso_btn.set_toggle(False)
        self._subplotso_btn.set_button_title('spacing options')
        self._subplotso_btn.valueChanged.connect(self.subplotSpacingOptions)
        self._collapsables.append(self._subplotso_btn)
        self.adj_window = None

        plot_options_layout = QtWidgets.QHBoxLayout()
        plot_options_layout.addWidget(self._subplotso_btn)
        plot_options_layout.addWidget(self._lino_btn)

        grid_legend_lyt = QtWidgets.QHBoxLayout()
        grid_legend_lyt.addWidget(self._legend_btn)
        grid_legend_lyt.addWidget(self._grid_btn)

        autoscale_scale_lyt = QtWidgets.QHBoxLayout()
        autoscale_scale_lyt.addWidget(self._autoscale_btn)
        autoscale_scale_lyt.addWidget(self._xscale_btn)
        autoscale_scale_lyt.addWidget(self._yscale_btn)
        autoscale_scale_lyt.addWidget(self._xeq0_btn)
        autoscale_scale_lyt.addWidget(self._yeq0_btn)

        # HLINES
        self._hline1 = QtWidgets.QFrame()
        self._hline1.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Sunken)
        self._hline1.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self._collapsables.append(self._hline1)

        self._hline2 = QtWidgets.QFrame()
        self._hline2.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Sunken)
        self._hline2.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self._collapsables.append(self._hline2)

        self._hline3 = QtWidgets.QFrame()
        self._hline3.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Sunken)
        self._hline3.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self._collapsables.append(self._hline3)

        spc = 10
        self._spacer1 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._spacer2 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._spacer3 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._spacer4 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._spacer5 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._spacer6 = QtWidgets.QSpacerItem(1,spc,QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._collapsables.append(self._spacer1)
        self._collapsables.append(self._spacer2)
        self._collapsables.append(self._spacer3)
        self._collapsables.append(self._spacer4)
        self._collapsables.append(self._spacer5)
        self._collapsables.append(self._spacer6)

        # panel layout
        vbox.addLayout(plotlabels)

        vbox.addSpacerItem(self._spacer1)
        vbox.addWidget(self._hline1)
        vbox.addSpacerItem(self._spacer2)

        vbox.addLayout(lims)
        #vbox.addLayout(scale_options_layout)
        #vbox.addWidget(self._autoscale_btn)
        vbox.addLayout(autoscale_scale_lyt)

        vbox.addSpacerItem(self._spacer3)
        vbox.addWidget(self._hline2)
        vbox.addSpacerItem(self._spacer4)

        vbox.addLayout(ticks)
        #vbox.addWidget(self._legend_btn)
        vbox.addLayout(grid_legend_lyt)
        vbox.addLayout(plot_options_layout)
        #vbox.addWidget(self._lino_btn)
        #vbox.addWidget(self._subplotso_btn)

        vbox.addSpacerItem(self._spacer5)
        vbox.addWidget(self._hline3)
        vbox.addSpacerItem(self._spacer6)

        vbox.addWidget(self._hold_btn)
        # vbox.addWidget(self._origin_axes_btn)

        vbox.insertStretch(-1,1)
        vbox.addWidget(self._reset_btn)

        # plot window
        self._data = None
        self._plotwindow = self.create_main_frame()

        # put side panel and plot window together
        hbox = QtWidgets.QHBoxLayout()
        hbox.addLayout(vbox)
        hbox.addLayout(self._plotwindow)
        hbox.setStretch(0,0)
        hbox.setStretch(1,11)
        self.setLayout(hbox)

        #self._on_draw() # draw once to get initial settings
        #self.copySubplotSettings()

        # Don't hide side-panel options by default
        self.set_collapsed(False)
        self.set_grid(True)
        self.set_autoscale(True)

        # DEFAULTS
        self._init_parms_()

    # setters
    def set_val(self, data):
        '''Takes a list of npy arrays.
        '''
        if isinstance(data, list):
            self._data = data
            self.on_draw()

    def set_grid(self, val):
        self._grid_btn.set_val(val)
        self.on_draw()

    def set_autoscale(self, val):
        self._autoscale_btn.set_val(val)
        self.on_draw()

    def set_collapsed(self, val):
        """bool | Only collapse the display options, not the Plot window.
        """
        self._isCollapsed = val
        for wdg in self._collapsables:
            if hasattr(wdg, 'setVisible'):
                wdg.setVisible(not val)

    def set_xlim(self, val, quiet=False):
        '''tuple of floats (min, max)
        '''
        self._xh.set_val(val[0])
        self._xl.set_val(val[1])
        if not quiet:
            self.on_draw()

    def set_ylim(self, val, quiet=False):
        '''tuple of floats (min, max)
        '''
        self._yh.set_val(val[0])
        self._yl.set_val(val[1])
        if not quiet:
            self.on_draw()

    def set_plotOptions(self, val):
        self._subplotSettings = val

    def set_lineOptions(self, val):
        self._lineSettings = val

    def set_plotPosition(self, val):
        self._subplotPosition = val

    def set_ticks(self, s):
        self._x_numticks.set_val(s['xticknum'])
        self._y_numticks.set_val(s['yticknum'])
        self._x_ticks.setText(s['xticks'])
        self._y_ticks.setText(s['yticks'])

    def set_plotlabels(self, s):
        self._plot_title.setText(s['title'])
        self._plot_xlab.setText(s['xlabel'])
        self._plot_ylab.setText(s['ylabel'])
        self.on_draw()

    def set_legend(self, val):
        self._legend_btn.set_val(val)
        self.on_draw()

    def set_xline(self, val):
        self._yeq0_btn.set_val(val)
        self.on_draw()

    def set_yline(self, val):
        self._xeq0_btn.set_val(val)
        self.on_draw()

    def set_scale(self, val):
        self._xscale_btn.set_val(val['xscale'])
        self._yscale_btn.set_val(val['yscale'])
        self.on_draw()

    # getters
    def get_val(self):
        return self._data

    def get_grid(self):
        return self._grid_btn.get_val()

    def get_autoscale(self):
        return self._autoscale_btn.get_val()

    def get_xlim(self):
        return (self._xh.get_val(), self._xl.get_val())

    def get_ylim(self):
        return (self._yh.get_val(), self._yl.get_val())

    def get_plotOptions(self):
        return self._subplotSettings

    def get_lineOptions(self):
        return self._lineSettings

    def get_plotPosition(self):
        return self._subplotPosition

    def get_ticks(self):
        s = {}
        s['xticknum'] = self._x_numticks.get_val()
        s['yticknum'] = self._y_numticks.get_val()
        s['xticks'] = str(self._x_ticks.text())
        s['yticks'] = str(self._y_ticks.text())
        return s

    def get_plotlabels(self):
        s = {}
        s['title'] = str(self._plot_title.text())
        s['xlabel'] = str(self._plot_xlab.text())
        s['ylabel'] = str(self._plot_ylab.text())
        return s

    def get_legend(self):
        return self._legend_btn.get_val()

    def get_xline(self):
        return self._yeq0_btn.get_val()

    def get_yline(self):
        return self._xeq0_btn.get_val()

    def get_scale(self):
        s = {}
        s['xscale'] = self._xscale_btn.get_val()
        s['yscale'] = self._yscale_btn.get_val()
        return s

    # support
    def check_validticks(self, tickwdg):
        s = tickwdg.text()
        comma_cnt = s.count(',')
        if (comma_cnt > 0) or (len(s) == 0):
            color = '#ffffff' # white
            tickwdg.setStyleSheet('QLineEdit { background-color: %s }' % color)
            return

        #color = '#fff79a' # yellow
        color = '#f6989d' # red
        tickwdg.setStyleSheet('QLineEdit { background-color: %s }' % color)
        return

    def create_main_frame(self):
        self.fig = Figure((6.0, 4.8), dpi=100, facecolor='0.98', linewidth=6.0, edgecolor='0.93')
        self.axes = None
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setParent(self)
        self.canvas.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.canvas.setFocus()

        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)

        #self.mpl_toolbar = NavigationToolbar(self.canvas, self)
        self.mpl_toolbar = NavbarTools(self.canvas, self)
        self.mpl_toolbar.actionTriggered.connect(self.copySubplotSettings)

        self.canvas.mpl_connect('key_press_event', self.on_key_press)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.canvas)  # the matplotlib canvas
        vbox.addWidget(self.mpl_toolbar)
        return vbox
        #self.setLayout(vbox)

    def lineOptionsDialog(self):
        if self.axes is None:
            print("Matplotlib: no lines to modify, skipping line editor")
            return
        figure_edit(self.axes, self)
        self.copySubplotSettings()
        self.on_draw()

    def subplotSpacingOptions(self):
        if self.fig is None:
            return

        # don't allow the user to open extra windows
        if self.adj_window is not None:
            if self.adj_window.isActive():
                self.adj_window.raise_()
                return

        self.adj_window = MainWin_close()
        self.adj_window.window_closed.connect(self.copySubplotSettings)
        win = self.adj_window

        win.setWindowTitle("Subplot Configuration Tool")
        image = os.path.join( matplotlib.rcParams['datapath'],'images','matplotlib.png' )
        win.setWindowIcon(QtGui.QIcon( image ))

        tool = SubplotToolQt(self.fig, win)
        win.setCentralWidget(tool)
        win.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        win.show()

    def copySubplotSettings(self):
        '''Get a copy of the settings found in the 'Figure Options' editor.
        '''
        if self.axes is None:
            return

        # subplot settings
        for k in self._subplot_keepers:
            self._subplotSettings[k] = getattr(self.axes, 'get_'+k)()

        # line settings
        self._lineSettings = []
        for l in self.axes.get_lines():
            s = {}
            for k in self._line_keepers:
                s[k] = getattr(l, 'get_'+k)()
            self._lineSettings.append(s)

        # subplot position
        self._subplotPosition = {}
        self._subplotPosition['left'] = self.fig.subplotpars.left
        self._subplotPosition['right'] = self.fig.subplotpars.right
        self._subplotPosition['top'] = self.fig.subplotpars.top
        self._subplotPosition['bottom'] = self.fig.subplotpars.bottom
        self._subplotPosition['wspace'] = self.fig.subplotpars.wspace
        self._subplotPosition['hspace'] = self.fig.subplotpars.hspace

    def applySubplotSettings(self):
        '''Everytime the plot is drawn it looses its 'Figure Options' so just
        make sure they are applied again.
        '''
        # subplot settings
        for k in self._subplot_keepers:
            if k in self._subplotSettings:
                getattr(self.axes, 'set_'+k)(self._subplotSettings[k])

        #subplot position
        self.fig.subplots_adjust(**self._subplotPosition)

    def _init_parms_(self):
        '''Default parameter settings
        '''
        self._subplotSettings = {}
        self._subplotPosition = {'right': 0.913, 'bottom': 0.119, 'top': 0.912, 'wspace': 0.2, 'hspace': 0.2, 'left': 0.111}
        self._lineSettings = []
        self.set_autoscale(True)
        self.set_grid(True)
        s = {}
        s['xticknum'] = 5
        s['yticknum'] = 5
        s['xticks'] = ''
        s['yticks'] = ''
        self.set_ticks(s)
        s = {}
        s['title'] = ''
        s['xlabel'] = ''
        s['ylabel'] = ''
        self.set_plotlabels(s)
        self.set_legend(False)
        s = {}
        s['xscale'] = False # linear
        s['yscale'] = False
        self.set_scale(s)
        self.on_draw()

    def on_draw(self):
        self._on_draw_cnt += 1
        if not self._updatetimer.isActive():
            self._updatetimer.start()

    def _on_draw(self):

        # HOLD / Create New AXES
        if not self._hold_btn.get_val():
            self.fig.clear()
            self.axes = self.fig.add_subplot(111)

        # AUTOSCALE and LIMITS
        self.axes.set_autoscale_on(self.get_autoscale())
        if not self.get_autoscale():
            self.axes.set_xlim(self.get_xlim())
            self.axes.set_ylim(self.get_ylim())

        # TITLE, XLABEL and YLABEL
        self.axes.set_title(self.get_plotlabels()['title'], fontweight='bold', fontsize=16)
        self.axes.set_xlabel(self.get_plotlabels()['xlabel'], fontsize=14)
        self.axes.set_ylabel(self.get_plotlabels()['ylabel'], fontsize=14)

        # self.axes.plot(self.x, self.y, 'ro')
        # self.axes.imshow(self.data, interpolation='nearest')
        # self.axes.plot([1,2,3])

        # XSCALE
        if self.get_scale()['xscale']:
            self.axes.set_xscale('log')
        else:
            self.axes.set_xscale('linear')

        # YSCALE
        if self.get_scale()['yscale']:
            self.axes.set_yscale('log')
        else:
            self.axes.set_yscale('linear')

        # GRID
        ax_color = '0.5'
        if self.get_grid():
            self.axes.grid(self.get_grid(), color=ax_color)
        else:
            self.axes.grid(self.get_grid())

        # AXES SPINE COLOR
        self.axes.spines['bottom'].set_color(ax_color)
        self.axes.spines['top'].set_color(ax_color)
        self.axes.spines['right'].set_color(ax_color)
        self.axes.spines['left'].set_color(ax_color)
        try:
            # deprecated in Matplotlib 2.0
            self.axes.set_axis_bgcolor('0.97')
        except AttributeError:
            self.axes.set_facecolor('0.97')

        # if self._origin_axes_btn.get_val():
        #     self.axes.spines['left'].set_position('zero')
        #     self.axes.spines['bottom'].set_position('zero')
        #     self.axes.spines['left'].set_smart_bounds(True)
        #     self.axes.spines['bottom'].set_smart_bounds(True)
        #     self.axes.xaxis.set_ticks_position('bottom')
        #     self.axes.yaxis.set_ticks_position('left')

        if self._data is None:
            return

        try:
            self.fig.hold(True)
        except:
            pass

        # plot each set
        # print "--------------------plot the data"
        for data in self._data:
            ln = max(data.shape)
            lw = max(5.0-np.log10(ln), 1.0)
            if ln > 0:
                al = max(1.0-1.0/np.log2(ln), 0.75)
            else:
                al = 0

            if data.shape[-1] == 2:
                self.axes.plot(data[..., 0], data[..., 1], alpha=al, lw=lw)
            else:
                self.axes.plot(data, alpha=al, lw=lw)

        # X=0, Y=0
        if self.get_xline():
            self.axes.axhline(y=0, color='k', zorder=-1, label="y=0")
        if self.get_yline():
            self.axes.axvline(x=0, color='k', zorder=-1, label="x=0")

        # LEGEND
        if self.get_legend():
            handles, labels = self.axes.get_legend_handles_labels()
            self.axes.legend(handles, labels)

        # AUTOSCALE
        if self.get_autoscale():
            self.set_xlim(self.axes.get_xlim(), quiet=True)
            self.set_ylim(self.axes.get_ylim(), quiet=True)

        # X TICKS
        xl = self._x_ticks.text().split(',')
        if len(xl) > 1:
            self.axes.set_xticks(np.linspace(*self.axes.get_xlim(), num=len(xl)))
            self.axes.set_xticklabels(xl)
        else:
            self.axes.set_xticks(np.linspace(*self.axes.get_xlim(), num=self._x_numticks.get_val()))

        # Y TICKS
        yl = self._y_ticks.text().split(',')
        if len(yl) > 1:
            self.axes.set_yticks(np.linspace(*self.axes.get_ylim(), num=len(yl)))
            self.axes.set_yticklabels(yl)
        else:
            self.axes.set_yticks(np.linspace(*self.axes.get_ylim(), num=self._y_numticks.get_val()))

        self.applySubplotSettings()

        self.canvas.draw()

        #print 'draw count: ', self._on_draw_cnt
        self._on_draw_cnt = 0

    def on_key_press(self, event):
        # print 'Matplotlib-> you pressed:' + str(event.key)
        # implement the default mpl key press events described at
        # http://matplotlib.org/users/navigation_toolbar.html#navigation-
        # keyboard-shortcuts
        try:
            from matplotlib.backend_bases import key_press_handler
            key_press_handler(event, self.canvas, self.mpl_toolbar)
        except:
            print("key_press_handler import failed. -old matplotlib version.")


class ExternalNode(gpi.NodeAPI):

    """A Qt embedded plot window originally using the code from:
    http://matplotlib.org/examples/user_interfaces/embedding_in_qt4_wtoolbar.html
    Updated by DDB in Feb. 2020 using the qt5 code from:
    https://matplotlib.org/examples/user_interfaces/embedding_in_qt5.html
    keyboard shortcuts can be found here:
    http://matplotlib.org/users/navigation_toolbar.html#navigation-keyboard-shortcuts

    INPUTS
    Up to 8 data sets can be plotted simultaneously
      1D real-valued data are plotted as graph
      2D data where the 2nd dimension is 2 will be plotted as X-Y parametric plot, otherwise
      all other 2D data are plotted as series of 1D plots
    """

    def initUI(self):
        # Widgets
        self.addWidget('MatplotDisplay', 'Plot')

        # IO Ports
        self.inport_range = list(range(0, 8))
        for i in self.inport_range:
            self.addInPort('in' + str(i), 'NPYarray', drange=(1,2), obligation=gpi.OPTIONAL)

    def compute(self):

        # check input ports for data
        in_lst = [self.getData('in' + str(i))
                  for i in self.inport_range if self.getData('in' + str(i)) is not None]

        self.setAttr('Plot', val=in_lst)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_APPLOOP



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Aug

import gpi
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """Computes Sampling Density Correction for 2D and 3D waveforms

    INPUTS
    crds - input coordinates, which range from -0.5 to +0.5.
           the last dimension is 2 or 3, corresponding to 2D (kx/ky) or 3D (kx/ky/kz)
    wates - optional relative weights, used to preferentially use some data over others in areas of overlap
    params_in - optional dictionary from (e.g.) spiralcoords, to automatically specify the effective matrix

    OUTPUTS - sampling density, same size as crds (minus the last dimension)

    WIDGETS:
    computenow - turn off to change effective matrix without computing the sampling density
    Dims per Set - how many of the dimensions to compute together as 1 set of coordinates.  Allows for extra
                   dimensions for (e.g.) slices, diffusion weightings, whatever
    Iterations - times to iterate algorithm
    Effective Matrix XY - FOV/resolution in X&Y, which indicates the width of data correlation in k-space
    Effective Matrix Z - FOV/resolution in Z, which indicates the width of data correlation in k-space
    Taper - taper the weights at the edge of k-space; value indicates the fraction of k-space radius to
            (linearly) taper from 1 to 0 (at the very edge): 0 gives no taper, 1 is "full" taper
    """

    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):      

        # Widgets
        self.addWidget('PushButton','computenow',toggle=True)
        self.addWidget('Slider','Dims per Set',min=1)
        self.addWidget('SpinBox','Iterations',val=1, min=1)
        self.addWidget('DoubleSpinBox','Effective MTX XY',val=300.0, min=2.0)
        self.addWidget('DoubleSpinBox','Effective MTX Z',val=300.0, min=2.0)
        self.addWidget('DoubleSpinBox','Taper',val=0.0, min=0.0, max = 1.0, singlestep = 0.01)
        self.addWidget('DoubleSpinBox','krad Scale',val=1.0, min=0.2, max = 2.0, singlestep = 0.1)

        # IO Ports
        self.addInPort('crds', 'NPYarray')
        self.addInPort('wates', 'NPYarray', obligation = gpi.OPTIONAL)
        self.addInPort('params_in', 'DICT', obligation = gpi.OPTIONAL)
        self.addOutPort('sdc', 'NPYarray')

    def validate(self):
        crds  = self.getData('crds')
        wates  = self.getData('wates')
        inparam = self.getData('params_in')

        # Check for compatibility
        if crds.shape[-1] not in [1,2,3]:
          self.log.warn("SDC(): Warning, data is not 1-, 2-, or 3-vec, skipping...")
          return 1
        if wates is not None:
          if wates.shape != crds[...,0].shape:
            self.log.warn("Wates Dimension does not match coords")
            return 1

        # Set a couple of conditional attributes
        self.setAttr('Effective MTX Z', visible=crds.shape[-1]==3)
        self.setAttr('Dims per Set', max = crds.ndim-1)

        # Auto Matrix calculation: extra 25% assumes "true resolution"
        # A more robust version will check to see if this is spiral, etc...
        if (inparam is not None):
            if 'headerType' in inparam:
                # old ReadPhilips param output
                if inparam['headerType'] == 'BNIspiral':
                    mtx_xy = (1.25*float(inparam['spFOVXY'][0])/
                              float(inparam['spRESXY'][0]))
                    stype = int(float(inparam['spSTYPE'][0]))
                    if crds.shape[-1] == 3:
                        mtx_z  = (float(inparam['spFOVZ'][0])/
                                  float(inparam['spRESZ'][0]))
                        if stype in [2,3]:  # SDST, FLORET
                            mtx_z *= 1.25
                        self.setAttr('Effective MTX Z', val=mtx_z)

                # new ReadPhilips param output
                elif inparam['headerType'] == 'spparams':
                    stype = inparam['SPIRAL_TYPE']
                    mtx_xy = (1.25*inparam['FOV_CM'][0] / inparam['RES_CM'][0])
                    if crds.shape[-1] == 3:
                        mtx_z = (inparam['FOV_CM'][2] / inparam['RES_CM'][2])
                        if stype in [2,3]:  # SDST, FLORET
                            mtx_z *= 1.25
                        self.setAttr('Effective MTX Z', val=mtx_z)
                else:
                    self.log.warn("wrong header type")
                    return 1
            else:
                # if there is no header type, then its also the wrong type
                self.log.warn("wrong header type")
                return 1

            self.setAttr('Effective MTX XY', val = mtx_xy)

    def compute(self):

        import numpy as np

        dps = self.getVal('Dims per Set')
        mtx_xy = self.getVal('Effective MTX XY')
        mtx_z  = self.getVal('Effective MTX Z')
        numiter = self.getVal('Iterations')
        taper = self.getVal('Taper')
        kradscale = self.getVal('krad Scale')
        crds   = self.getData('crds')
        inwates  = self.getData('wates')

        mtxsz_xy = (2.*mtx_xy)+6
        mtxsz_z  = (2.*mtx_z)+6

        sdshape = crds[...,0].shape
        maxi = crds.ndim - 2
        npts = 1
        for i in range(dps):
          npts *= sdshape[maxi-i]
        nsets = int(crds[...,0].size/npts)

        # Reshape crds to be 3 dimensional - # sets, # pts/set, and dimensionality (1-3)
        crds = np.reshape(crds.astype(np.float64),(nsets,npts,crds.shape[-1]))

        if self.getVal('computenow'):

          if inwates is not None:
            wates = np.copy(inwates.reshape(nsets,npts).astype(np.float64))
          else:
            wates = np.ones((nsets,npts), dtype=np.float64)

          # import in thread to save namespace 
          import gpi_core.gridding.sdc as sd
          for set in range(nsets):
            if crds.shape[-1] == 1:
              cmtxdim = np.array([mtxsz_xy],dtype=np.int64)
              sdcset = sd.oned_sdc(crds[set,:],wates[set,:],cmtxdim,numiter,taper)
            if crds.shape[-1] == 2:
              cmtxdim = np.array([mtxsz_xy,mtxsz_xy],dtype=np.int64)
              sdcset = sd.twod_sdc(crds[set,:],wates[set,:],cmtxdim,numiter,taper)
            if crds.shape[-1] == 3:
              cmtxdim = np.array([mtxsz_xy,mtxsz_xy,mtxsz_z],dtype=np.int64)
              sdcset = sd.threed_sdc(crds[set,:],wates[set,:],cmtxdim,numiter,taper,kradscale)
            if set == 0:
              sdc = np.expand_dims(sdcset,0)
            else:
              sdc = np.append(sdc,np.expand_dims(sdcset,0),axis=0)

          # Reshape sdc weights to match that of incoming coordinates
          self.setData('sdc', np.reshape(sdc,sdshape))

        return(0)




# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Payal Bhavsar
# Date: 2013Dec
# Mods: Jim Pipe, 2013 Dec

import numpy as np
import gpi
import math
import scipy.ndimage as scnd

class ExternalNode(gpi.NodeAPI):

    """A node to rotate data about a plane specified by two axes

    INPUT: input numpy array

    OUTPUT: output numpy array

    WIDGETS:
    Theta is entered in degrees
    Output can be reshaped to contain the input after rotation
    Modes of rotation - constant, nearest, reflect and wrap can be chosen
    Order for interpolation can be specified.

    """

    def initUI(self):
        self.addInPort('in', 'NPYarray')
        self.addOutPort('out', 'NPYarray')

        self.addWidget('DoubleSpinBox', 'Theta', val=0.0, decimals = 5)
        self.addWidget('ExclusivePushButtons', 'Units', buttons=['Deg','Rad','Cyc'], val = 0)
        self.dim_buttons = ['[0-1]', '[0-2]', '[0-3]', '[0-4]', '[0-5]', '[1-2]', '[1-3]', '[1-4]', '[1-5]', '[2-3]','[2-4]','[2-5]','[3-4]','[3-5]','[4-5]']
        self.addWidget('ExclusivePushButtons', 'Plane of Rotation:', buttons=self.dim_buttons, val=0)
        self.addWidget('PushButton', 'Reshape', toggle=True, val=0, collapsed = True)
        self.op_buttons_1 = ['constant','nearest','reflect','wrap']
        self.addWidget('ExclusiveRadioButtons','Mode', buttons=self.op_buttons_1, val=0, collapsed = True)
        self.addWidget('Slider', 'Order', val = 1, min=0, max=5, collapsed = True)

    def validate(self):
        data = self.getData('in')
        self.setAttr('Theta', max=1000., min=-1000.0)
        ddim = (data.ndim)

        if ('in' in self.portEvents()):
          if ddim == 2 or self.getVal('Plane of Rotation:') > ddim-2:
            self.setAttr('Plane of Rotation:', buttons=self.dim_buttons, val=0)
        if ddim == 2:
          self.dim_buttons = ['[0-1]']
        if ddim == 3:
          self.dim_buttons = ['[0-1]','[1-2]','[0-2]']
        if ddim == 4:
          self.dim_buttons = ['[0-1]','[1-2]', '[0-2]', '[1-3]', '[2-3]', '[0-3]']
        if ddim == 5:
          self.dim_buttons = ['[0-1]', '[1-2]', '[0-2]', '[1-3]', '[2-3]', '[0-3]','[0-4]', '[1-4]', '[2-4]', '[3-4]']
        op = self.getVal('Plane of Rotation:')
        self.setAttr('Plane of Rotation:', buttons=self.dim_buttons, val=op)

        return(0)

    def compute(self):
        import numpy as np

        units = self.getVal('Units')
        if units == 0: # degrees
            unitConv = 1
        elif units == 1: # radians
            unitConv = 180/np.pi
        else: # cycles
            unitConv = 360

        data = self.getData('in')
        Theta = self.getVal('Theta') * unitConv
        valueorder = self.getVal('Order')
        value = self.getVal('Reshape')
        self.shape = list(data.shape)
        op = self.getVal('Plane of Rotation:')
        mode_val = self.getVal('Mode')


        if op == 0:
            ax0 = 0
            ax1 = 1

        if op == 1:
            ax0 = 1
            ax1 = 2

        if op == 2:
            ax0 = 0
            ax1 = 2

        if op == 3:
            ax0 = 1
            ax1 = 3

        if op == 4:
            ax0 = 2
            ax1 = 3

        if op == 5:
            ax0 = 0
            ax1 = 3

        if op == 6:
            ax0 = 0
            ax1 = 4

        if op == 7:
            ax0 = 1
            ax1 = 4

        if op == 8:
            ax0 = 2
            ax1 = 4

        if op == 9:
            ax0 = 3
            ax1 = 4

        if np.iscomplexobj(data):
          if mode_val == 0:
            RD = scnd.rotate(data.real,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='constant')
            ID = scnd.rotate(data.imag,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='constant')
          elif mode_val == 1:
            RD = scnd.rotate(data.real,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='nearest')
            ID = scnd.rotate(data.imag,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='nearest')
          elif mode_val == 2:
            RD = scnd.rotate(data.real,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='reflect')
            ID = scnd.rotate(data.imag,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='reflect')
          elif mode_val == 3:
            RD = scnd.rotate(data.real,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='wrap')
            ID = scnd.rotate(data.imag,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='wrap')
          out = RD + 1j*ID
        else:
          if mode_val == 0:
            out = scnd.rotate(data,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='constant')
          elif mode_val == 1:
            out = scnd.rotate(data,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='nearest')
          elif mode_val == 2:
            out = scnd.rotate(data,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='reflect')
          elif mode_val == 3:
            out = scnd.rotate(data,Theta,(ax0,ax1), reshape=value, order = valueorder, mode='wrap')

        self.setData('out', out.astype(data.dtype))

        return(0)



# Author: Ashley Anderson
# Date: 2015-12-18 13:36

import gpi
import os
import shutil
import tempfile
import urllib.request

class ExternalNode(gpi.NodeAPI):
    """This node will download a file and save it to the disk. If the file is a
    recognized format, the data can then be read into GPI using one of the
    readers in core.fileIO.

    OUTPUT:
        path - local filepath for the downloaded file
    WIDGETS:
        url - the URL of the file you want to download
        save local copy - if enabled, you can choose the path to save the file,
            otherwise it will be saved to a temp file
        save file - where to save the file
        re-download - download the file again, even if it is stored locally
    """

    # initialize the UI - add widgets and input/output ports
    def initUI(self):
        # Widgets
        self.addWidget('StringBox', 'url', placeholder="https://www.gpilab.com/example_data.npy")
        self.addWidget('PushButton', 'save local copy', toggle=True, val=False)
        self.addWidget('SaveFileBrowser', 'save file', button_title='Browse')
        self.addWidget('PushButton', 're-download')

        self.addWidget('PushButton', 'validate', toggle=True, val=False, visible=False)
        self.addWidget('StringBox', 'checksum', visible=False)

        # IO Ports
        self.addOutPort('path', 'STRING')

        self.url = None

    # validate the data - runs immediately before compute
    # your last chance to show/hide/edit widgets
    # return 1 if the data is not valid - compute will not run
    # return 0 if the data is valid - compute will run
    def validate(self):
        if self.getVal('save local copy'):
            self.setAttr('save file', visible=True)
        else:
            self.setAttr('save file', visible=False)

        # don't overwrite an existing file
        if os.path.exists(self.getVal('save file')):
            self.log.error("For safety, this node will not overwrite existing "
                           + "files. ")
            self.log.error("If you really want to overwrite the file "
                           + self.getVal('save file')
                           + " please delete it manually.")
            return 1

        # clear the path if the URL has changed
        # this will cause a re-download when the URL is changed
        if self.url != self.getVal('url'):
            self.setData('path', None)
            self.url = self.getVal('url')

        if self.url:
            self.setDetailLabel(self.url)

        return 0

    # process the input data, send it to the output port
    # return 1 if the computation failed
    # return 0 if the computation was successful
    def compute(self):
        # logic regarding when to skip the download
        if not (self.getData('path') is None
                or 're-download' in self.widgetEvents()):
            if self.getVal('save local copy') and 'save file' in self.widgetEvents():
                try:
                    shutil.copyfile(self.getData('path'), self.getVal('save file'))
                    self.setData('path', self.getVal('save file'))
                except OSError as e:
                    self.log.error("Couldn't copy temp file to "
                                   + self.getVal('save file')
                                   + " Try re-downloading.")
            return 0

        # prevent the node from running when first instantiated
        if self.url is None or self.url == '':
            return 0

        self.log.info("Downloading file from: {}".format(self.url))
        try:
            R = urllib.request.urlopen(self.url)
        except ValueError as e:
            self.log.error("Unknown URL type."
                           + "Make sure it starts with the protocol"
                           + "e.g. 'http://'.")
            return 1

        fp = None
        if self.getVal('save local copy'):
            try:
                fp = open(self.getVal('save file'), 'w+b')
            except FileNotFoundError:
                self.log.warn("File path " + self.getVal('save file')
                              + "not available. "
                              +"Saving to a temporary file instead.")
        if fp is None:
            fp = tempfile.NamedTemporaryFile(delete=False)

        # here is where the data is actually downloaded (may take a while)
        fp.write(R.read())
        fp.close()

        out = fp.name
        self.setData('path', out)

        return 0

    def execType(self):
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2012sep02

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Output the Real, Imaginary, Magnitude, or Phase of a Complex array, element-wise
    INPUT - complex data array

    OUTPUT - real-valued array

    WIDGETS:
    R I M P - select Real, Imaginary, Magnitude, or Phase
    Units - for Phase, select Degrees, Radians, or Cycles
    Unwrap - for Phase, you can do simple phase unwrapping in one direction
    Unwrap Dimension - for Phase, choose the dimension for phase unwrapping
    """

    def initUI(self):

        # Widgets
        self.addWidget('ExclusivePushButtons', 'R I M P', buttons=[
                       'R', 'I', 'M', 'P'], val=2)
        self.addWidget('ExclusivePushButtons', 'Units', buttons=[
                       'Deg', 'Rad', 'Cyc'], val=1)
        self.addWidget('PushButton', 'UnWrap', toggle=True)
        self.addWidget('Slider', 'UnWrap Dimension', val = 0, min=0)

        # IO Ports
        self.addInPort('in', 'NPYarray', dtype=[np.complex64, np.complex128],
                       obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray', dtype=[np.float32, np.float64])

        # operations
        self.op = [np.real, np.imag, np.abs, np.angle]

    def validate(self):
        data = self.getData('in')

        self.setAttr('UnWrap Dimension', max=(len(data.shape)-1), min=(-len(data.shape)))

        if self.getVal('R I M P') == 3:
          self.setAttr('Units',visible=True)
          self.setAttr('UnWrap',visible=True)
          self.setAttr('UnWrap Dimension',visible=self.getVal('UnWrap'))
        else:
          self.setAttr('Units',visible=False)
          self.setAttr('UnWrap',visible=False)
          self.setAttr('UnWrap Dimension',visible=False)

        L = ['Real', 'Imaginary', 'Magnitude', 'Phase']
        self.setDetailLabel(L[self.getVal('R I M P')])

        return(0)

    def compute(self):

        import numpy as np

        data = self.getData('in')

        out = self.op[self.getVal('R I M P')](data)
        if self.getVal('R I M P') == 3:
          if self.getVal('UnWrap'):
            out = np.unwrap(out, discont = np.pi, axis = self.getVal('UnWrap Dimension'))
          if self.getVal('Units') == 0:
            out = 180*out/np.pi
          if self.getVal('Units') == 2:
            out = out/(2.*np.pi)

        self.setData('out', out)
        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Jun13

import gpi
from gpi import QtWidgets


# WIDGET
class SpynAxys(gpi.GenericWidgetGroup):
    """A combination of SpinBoxes, DoubleSpinBoxes, and PushButtons
    to form a unique widget suitable for the Spin Generator Axes.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._val = {}
        self._val['length'] = 1
        self._val['start'] = 0.
        self._val['end'] = 0. # the original array length

        self.sb = gpi.BasicSpinBox()  # length
        self.sb.set_label('# Spins:')
        self.sb.set_min(1)
        self.sb.set_val(1)
        self.sb.set_max(gpi.GPI_INT_MAX)

        self.db1 = gpi.BasicDoubleSpinBox()  # start
        self.db1.set_label('Start:')
        self.db1.set_min(gpi.GPI_FLOAT_MIN)
        self.db1.set_max(gpi.GPI_FLOAT_MAX)
        self.db1.set_decimals(3)
        self.db1.set_singlestep(1.)
        self.db1.set_val(0)

        self.db2 = gpi.BasicDoubleSpinBox()  # end
        self.db2.set_label('End:')
        self.db2.set_min(gpi.GPI_FLOAT_MIN)
        self.db2.set_max(gpi.GPI_FLOAT_MAX)
        self.db2.set_decimals(3)
        self.db2.set_singlestep(1.)
        self.db2.set_val(0)

        self.sb.valueChanged.connect(self.lenChange)
        self.db1.valueChanged.connect(self.startChange)
        self.db2.valueChanged.connect(self.endChange)

        vbox = QtWidgets.QHBoxLayout()
        vbox.addWidget(self.sb)
        vbox.addWidget(self.db1)
        vbox.addWidget(self.db2)
        vbox.setStretch(0, 0)
        vbox.setStretch(1, 0)
        vbox.setStretch(2, 0)
        vbox.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        vbox.setSpacing(0)
        self.setLayout(vbox)

    # setters
    def set_val(self, val):
        """A python-dict containing: size, and width parms. """
        sig = False
        if 'length' in val:
            self._val['length'] = val['length']
            self.setLenQuietly(val['length'])
            sig = True
        if 'start' in val:
            self._val['start'] = val['start']
            self.setStartQuietly(val['start'])
            sig = True
        if 'end' in val:
            self._val['end'] = val['end']
            self.setEndQuietly(val['end'])
            sig = True
        if sig:
            self.valueChanged.emit()


    # getters
    def get_val(self):
        return self._val

    # support
    def lenChange(self, val):
        self._val['length'] = val
        self.setLenQuietly(self._val['length'])
        self.valueChanged.emit()

    def startChange(self, val):
        self._val['start'] = val
        self.setStartQuietly(self._val['start'])
        self.valueChanged.emit()

    def endChange(self, val):
        self._val['end'] = val
        self.setEndQuietly(self._val['end'])
        self.valueChanged.emit()

    def setLenQuietly(self, val):
        self.sb.blockSignals(True)
        self.sb.set_val(val)
        self.sb.blockSignals(False)

    def setStartQuietly(self, val):
        self.db1.blockSignals(True)
        self.db1.set_val(val)
        self.db1.blockSignals(False)

    def setEndQuietly(self, val):
        self.db2.blockSignals(True)
        self.db2.set_val(val)
        self.db2.blockSignals(False)


class ExternalNode(gpi.NodeAPI):
    """Module to generate initial spin profile, for bloch simulations in GPI
    Output data are 11-dimensional, with data stored as follows:
    Dimension 0 corresponds to base information such  M0, T1, T2, velocities, Frequency,
                Diffusion (not yet implemented), and the value of dt
    Dimension 1 corresponds to time.
                The first index of Dimension 1 stores the above info, and
                the 2nd index stores the Spin magnetization, position, time,
                                         gradient, rf, and spoiler info
    These two dimensions are illustrated below:

  D  ###### Dim 0 ->>> #########################
  i  #    0  1  2  3  4  5  6  7  8  9  10 11 12
  m  #
  1  # 0  M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
  |  # 1  Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
  v  # 2  Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
  v  # 3  Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
  v  # 4  Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
  v  ###########################################

    Dimensions 2-10 correspond to the widgets below, from T1, T2, etc. to Vy and Vz

    OUTPUT - spin state, typically fed to Bloch module.  No data present until "Starting Spins"
             widget is set to something other than "OFF"

    WIDGETS:
    Starting Spins - starting state of all spins.  When set to "OFF", no data are generated
    T0 (ms) - starting time.  Valuable for (e.g.) having velocities (which is not yet implemented)
    dt (us) - time between time points which will be recorded by Bloch module and pasted along the first dimension
              smaller numbers produce finer time resolution but larger data sets
              Note the actual numerical simulations occur with time resolution of 1us regardless of this value
    For the remaining widgets, the following is true:
      # Spins - length along that dimension, with values from the specified "Start" value to the specified "End" value
      if # Spins is 1, the value of that property is given by the "Start" parameter
    T1 (ms) - dimension 2 variable.  A value of 0 is infinity (no T1 recovery), otherwise values are in ms
    T2 (ms) - dimension 3 variable.  A value of 0 is infinity (no T2 relaxation), otherwise values are in ms
    Fq (Hz) - dimension 4 variable, off-resonance frequency
    X0/Y0/Z0 - dimension 5-7 variables, locations of spins
    Vx/Vy/Vz - dimension 8-10 variables, velocity of spins (not yet implemented)
    """

    def initUI(self):
        # Widgets
        self.addWidget('ExclusivePushButtons', 'Starting Spins',
                       buttons=['OFF', 'Mz', '0', '-Mz', 'Mx', 'My'],
                       val=0)
        self.addWidget('DoubleSpinBox', 'T0 (ms)', val=0.0)
        self.addWidget('SpinBox', 'dt (us)', val=50, min=1)
        self.addWidget('SpynAxys', 'T1 (ms)')
        self.addWidget('SpynAxys', 'T2 (ms)')
        self.addWidget('SpynAxys', 'Fq (Hz)')
        self.addWidget('SpynAxys', 'X0 (mm)')
        self.addWidget('SpynAxys', 'Y0 (mm)')
        self.addWidget('SpynAxys', 'Z0 (mm)')
        self.addWidget('SpynAxys', 'Vx (cm/s)')
        self.addWidget('SpynAxys', 'Vy (cm/s)')
        self.addWidget('SpynAxys', 'Vz (cm/s)')

        # IO Ports
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''

        # validate widget bounds
        return 0

    def compute(self):
       '''This is where the main algorithm should be implemented.
       '''

       import numpy as np

       # GETTING WIDGET INFO
       dim1 = 13 # number of params
         # 0  1  2  3  4  5  6  7  8  9  10 11 12
         # M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt __ __
         # Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
       dim2 = 2  # number of time points, 1st time point is constants

       dtms = float(self.getVal('dt (us)')) * 0.001 # convert from us to ms
       t0ms = self.getVal('T0 (ms)')

       mx=my=mz=0.
       if self.getVal('Starting Spins') == 1:
         mz = 1
       if self.getVal('Starting Spins') == 3:
         mz = -1
       if self.getVal('Starting Spins') == 4:
         mx = 1
       if self.getVal('Starting Spins') == 5:
         my = 1

       if self.getVal('Starting Spins') != 0: # OFF State

         # Units are ms, kHz, m, mT

         # Dim 3 is T1
         val = self.getVal('T1 (ms)')
         dim3 = val['length']

         if val['start']<=0:
           r1start=0
         else:
           r1start=1./val['start']

         if val['end']<=0:
           r1end=0
         else:
           r1end=1./val['end']

         r1step = (r1end-r1start)/max(1.,float(dim3-1))

         # Dim 4 is T2
         val = self.getVal('T2 (ms)')
         dim4 = val['length']

         if val['start']<=0:
           r2start=0
         else:
           r2start=1./val['start']

         if val['end']<=0:
           r2end=0
         else:
           r2end=1./val['end']

         r2step = (r2end-r2start)/max(1.,float(dim4-1))

         # Dim 5 is Fq, convert to kHz
         val = self.getVal('Fq (Hz)')
         dim5 = val['length']
         fqstart=0.001*val['start']
         fqstep= 0.001*(val['end']-val['start'])/max(1.,float(dim5-1))

         # Dim 6 is X0, convert to m
         val = self.getVal('X0 (mm)')
         dim6 = val['length']
         x0start=0.001*val['start']
         x0step= 0.001*(val['end']-val['start'])/max(1.,float(dim6-1))

         # Dim 7 is Y0, convert to m
         val = self.getVal('Y0 (mm)')
         dim7 = val['length']
         y0start=0.001*val['start']
         y0step= 0.001*(val['end']-val['start'])/max(1.,float(dim7-1))

         # Dim 8 is Z0, convert to m
         val = self.getVal('Z0 (mm)')
         dim8 = val['length']
         z0start=0.001*val['start']
         z0step= 0.001*(val['end']-val['start'])/max(1.,float(dim8-1))

         # Dim 9 is Vx, convert to m/ms
         val = self.getVal('Vx (cm/s)')
         dim9 = val['length']
         vxstart=0.00001*val['start']
         vxstep= 0.00001*(val['end']-val['start'])/max(1.,float(dim9-1))

         # Dim 10 is Vy, convert to m/ms
         val = self.getVal('Vy (cm/s)')
         dim10 = val['length']
         vystart=0.00001*val['start']
         vystep= 0.00001*(val['end']-val['start'])/max(1.,float(dim10-1))

         # Dim 11 is Vz, convert to m/ms
         val = self.getVal('Vz (cm/s)')
         dim11 = val['length']
         vzstart=0.00001*val['start']
         vzstep= 0.00001*(val['end']-val['start'])/max(1.,float(dim11-1))

         # Create the data
         out=np.zeros([dim1,dim2,dim3,dim4,dim5,dim6,dim7,dim8,dim9,dim10,dim11])

         # Assign the 1st element of data in dim2, M0 T1 T2 Vx Vy Vz Fq Dx Dy Dz dt GM __
         # These values don't change
         # M0
         out[0,0,:,:,:,:,:,:,:,:,:] = 1.

         # T1
         for i in range(0,dim3):
           out[1,0,i,:,:,:,:,:,:,:,:] = r1start + r1step*float(i)

         # T2
         for i in range(0,dim4):
           out[2,0,:,i,:,:,:,:,:,:,:] = r2start + r2step*float(i)

         #Vx
         for i in range(0,dim9):
           out[3,0,:,:,:,:,:,:,i,:,:] = vxstart + vxstep*float(i)

         #Vy
         for i in range(0,dim10):
           out[4,0,:,:,:,:,:,:,:,i,:] = vystart + vystep*float(i)

         #Vz
         for i in range(0,dim11):
           out[5,0,:,:,:,:,:,:,:,:,i] = vzstart + vzstep*float(i)

         #Fq
         for i in range(0,dim5):
           out[6,0,:,:,i,:,:,:,:,:,:] = fqstart + fqstep*float(i)

         # Dx, Dy, Dz = 0, no action for now

         # dt - convert from us to ms
         out[10,0,:,:,:,:,:,:,:,:,:] = dtms

         # Gamma - fixed for now, kHz/mT
         out[11,0,:,:,:,:,:,:,:,:,:] = 42.577

         # Assign 2nd element of data in dim2, Mx My Mz X  Y  Z  T  Gx Gy Gz Rx Ry Sp
         # This is the first time point

         #Mx My Mz
         out[0,1,:,:,:,:,:,:,:,:,:] = mx
         out[1,1,:,:,:,:,:,:,:,:,:] = my
         out[2,1,:,:,:,:,:,:,:,:,:] = mz

         #X0
         for i in range(0,dim6):
           out[3,1,:,:,:,i,:,:,:,:,:] = x0start + x0step*float(i)

         #Y0
         for i in range(0,dim7):
           out[4,1,:,:,:,:,i,:,:,:,:] = y0start + y0step*float(i)

         #Z0
         for i in range(0,dim8):
           out[5,1,:,:,:,:,:,i,:,:,:] = z0start + z0step*float(i)

         #T0
         out[6,1,:,:,:,:,:,:,:,:,:] = t0ms

         # Reset positions to reflect time relative to zero
         # X0 += Vx*T0
         out[3,1,:,:,:,:,:,:,:,:,:] += out[3,0,:,:,:,:,:,:,:,:,:]*t0ms
         # Y0 += Vy*T0
         out[4,1,:,:,:,:,:,:,:,:,:] += out[4,0,:,:,:,:,:,:,:,:,:]*t0ms
         # Z0 += Vz*T0
         out[5,1,:,:,:,:,:,:,:,:,:] += out[5,0,:,:,:,:,:,:,:,:,:]*t0ms

         #Gx Gy Gz Rx Ry Sp all 0, no action necessary

       else:
         out = None

       # SETTING PORT INFO
       self.setData('out', out)

       return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2013Aug

import gpi
import numpy as np


class ExternalNode(gpi.NodeAPI):
    """Implements rolloff correction in-house 2D gridding module written in C++.
    This module corrects the image shading created by the Grid module and crops the data by 1/3
    in each Gridded dimension to produce an image of the right matrix size

    INPUT: complex data - typically the output of Grid, Fourier Transformed to image space
           These data can represent 1D, 2D or 3D data, with optional extra dimensions representing (e.g.) coils or slices

    OUTPUT: complex data after Rolloff Correction

    WIDGETS:
    Num Rolloff Dims - Set to 1, 2, or 3 corresponding to 1D, 2D, or 3D gridded data sets
                       Remaining dims are treated independently, e.g. as slices, coils, etc.
    Isotropic FOV - multiplies data by a circular/spherical mask for 2D/3D data
    """

    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('Slider','Num Rolloff Dims',min=1,max=3,val=2)
        self.addWidget('PushButton','Isotropic FOV',toggle=True,val=True)

        # IO Ports
        self.addInPort('data', 'NPYarray', dtype=[np.complex64, np.complex128])
        self.addOutPort('out', 'NPYarray', dtype=[np.complex64, np.complex128])

    def validate(self):
        ''' DHW check if input data, coords and weight have the correct dimentions and sizes.
        '''

        data = self.getData('data')
        nrd = self.getVal('Num Rolloff Dims')

        if data.ndim < nrd:
          self.log.warn(" Error: Too few dimensions! ")
          return 1

        return(0)

    def compute(self):

        import numpy as np
        import gpi_core.gridding.grid as gd

        # force complex64 for underlying c++-routine
        data = self.getData('data')
        in_dtype = data.dtype  # make the out-type the same as the in-type

        dshape = np.array(data.shape)
        oshape = np.copy(dshape)
        nrd = self.getVal('Num Rolloff Dims')
        isofov = self.getVal('Isotropic FOV')

        # dimensions for rolloff - we will loop in this wrapper (below) around multiple sets
        # Note we are reversing the order of dimensions - necessary for python-c
        if nrd == 1:
          oshape[-1] = np.ceil(2*dshape[-1]/3.)
          outdim = np.array([oshape[-1] ],dtype=np.int64)
          nsets = np.prod(oshape)//np.prod(outdim)
          newdshape = np.array((nsets,dshape[-1]))
          outdim1 = outdim
        elif nrd == 2:
          oshape[-1] = np.ceil(2*dshape[-1]/3.)
          oshape[-2] = np.ceil(2*dshape[-2]/3.)
          outdim = np.array([oshape[-1], oshape[-2] ],dtype=np.int64)
          nsets = np.prod(oshape)//np.prod(outdim)
          newdshape = np.array((nsets,dshape[-2],dshape[-1]))
          outdim1 = outdim
        elif nrd == 3:
          oshape[-1] = np.ceil(2*dshape[-1]/3.)
          oshape[-2] = np.ceil(2*dshape[-2]/3.)
          oshape[-3] = np.ceil(2*dshape[-3]/3.)
          outdim = np.array([oshape[-1], oshape[-2], oshape[-3] ],dtype=np.int64)
          outdim1 = np.array([oshape[-3], oshape[-2], oshape[-1] ],dtype=np.int64)
          nsets = np.prod(oshape)//np.prod(outdim)
          newdshape = np.array((nsets,dshape[-3],dshape[-2],dshape[-1]))

        # Reshape data to have the first index be # sets, the rest be the rolled-off data
        data = np.reshape(data,newdshape)

        outshape1 = np.append([nsets], outdim1)
        out = np.zeros(outshape1,  dtype = in_dtype)

        for i in range(int(nsets)):
          # force single precision
          outset = gd.rolloff(data[i,...].astype(np.complex64),outdim,isofov)
          out[i,...] = np.require(outset, dtype=in_dtype)

        out = np.reshape(out, oshape)

        self.setData('out', np.require(out, dtype=in_dtype))

        return(0)




#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

import sys, os

# Check for Anaconda PREFIX, or assume that THIS file location is the CWD.
GPI_PREFIX = 'D:/bld/gpi_core_1583861226620/_h_env' # ANACONDA
if GPI_PREFIX == '/opt/' + 'anaconda1anaconda2anaconda3':
    GPI_PREFIX, _ = os.path.split(os.path.dirname(os.path.realpath(__file__)))
    GPI_LIB_DIR = os.path.join(GPI_PREFIX, 'lib')
    if GPI_LIB_DIR not in sys.path:
        sys.path.insert(0, GPI_LIB_DIR)

# gpi
from gpi import launch

if __name__ == '__main__':
    launch.launch()



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2013 July
# Brief: Quick readout of dictionary data types

import gpi
import numpy as np
import re
from gpi import QtWidgets
np.set_printoptions(linewidth=256)


# flatten a nested dictionary
def unwrap_dict(din, dout):
    for k in list(din.items()):
        if 'dict' in str(type(k[1])).lower():
            dout[k[0]] = k[1]
            unwrap_dict(din[k[0]], dout)
        else:
            dout[k[0]] = k[1]
    return dout


# add dictionary item to text output and handle slicing of lists and np arrays
def print_item(item, report, minl, maxl):
    if ('list' in str(type(item[1])) or 'ndarray' in str(type(item[1]))):
        report = report + str(item[0]) + ' = ' + str(item[1][minl:maxl]) + "\n"
    else:
        report = report + str(item[0]) + ' = ' + str(item[1]) + "\n"

    return report


def gen_keys_from_keywords(keys, indict, exactmatch):
    # add regex search results
    if keys != '':
        dflat = dict()
        dflat = unwrap_dict(indict, dflat)
        matching_keys = []
        keys = re.split(', | ,|,', keys)
        if exactmatch:
            for key in dflat:
                if key in keys:
                    matching_keys.append(key)
        else:
            for pattern in keys:
                cpat = re.compile(pattern.lower())
                for key in dflat:
                    if cpat.search(key.lower()):
                        matching_keys.append(key)
    else:
        matching_keys = []

    return(matching_keys)


class ReduceSliders(gpi.GenericWidgetGroup):
    """Combines the BasicCWFCSliders with ExclusivePushButtons
    for a unique widget element useful for reduce dimensions.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)
        self.sl = gpi.BasicCWFCSliders()
        self.sl.valueChanged.connect(self.valueChanged)
        # at least one button
        self.button_names = ['C/W', 'B/E', 'Slice', 'Pass']
        self.buttons = []
        cnt = 0
        wdgLayout = QtWidgets.QGridLayout()
        for name in self.button_names:
            newbutton = QtWidgets.QPushButton(name)
            newbutton.setCheckable(True)
            newbutton.setAutoExclusive(True)
            self.buttons.append(newbutton)
            newbutton.clicked.connect(self.findValue)
            newbutton.clicked.connect(self.valueChanged)
            wdgLayout.addWidget(newbutton, 0, cnt, 1, 1)
            cnt += 1
        # layout
        wdgLayout.addWidget(self.sl, 1, 0, 1, 4)
        wdgLayout.setVerticalSpacing(0)
        wdgLayout.setSpacing(0)
        self.setLayout(wdgLayout)
        # default
        self.set_min(1)
        self._selection = 3  # pass is default
        self.buttons[self._selection].setChecked(True)
        self.sl.set_allvisible(False)

    # setters
    def set_val(self, val):
        """A python-dict containing keys: center, width, floor, ceiling,
        and selection.
        """
        self.sl.set_center(val['center'])
        self.sl.set_width(val['width'])
        self.sl.set_floor(val['floor'])
        self.sl.set_ceiling(val['ceiling'])
        self._selection = val['selection']
        self.buttons[self._selection].blockSignals(True)
        self.buttons[self._selection].setChecked(True)
        self.buttons[self._selection].blockSignals(False)
        self.findValue(None)

    def set_min(self, val):
        """A min value for center, width, floor and ceiling (int)."""
        self.sl.set_min(val)

    def set_max(self, val):
        """A max value for center, width, floor and ceiling (int)."""
        self.sl.set_max(val)

    # getters
    def get_val(self):
        val = {}
        val['selection'] = self._selection
        val['center'] = self.sl.get_center()
        val['width'] = self.sl.get_width()
        val['floor'] = self.sl.get_floor()
        val['ceiling'] = self.sl.get_ceiling()
        return val

    def get_min(self):
        return self.sl.get_min()

    def get_max(self):
        return self.sl.get_max()
    # support

    def setCropBounds(self):
        self.sl.set_min_width(1)

    def setSliceBounds(self):
        self.sl.set_min_width(0)
        self.sl.set_width(1)

    def setPassBounds(self):
        self.sl.set_min_width(1)
        self.sl.set_width(self.get_max())
        # JGP For Pass, reset cwfc
        self.sl.set_center((self.get_max()+1)/2)
        self.sl.set_width(self.get_max())
        self.sl.set_floor(1)
        self.sl.set_ceiling(self.get_max())

    def findValue(self, value):
        cnt = 0
        for button in self.buttons:
            if button.isChecked():
                self._selection = cnt
            cnt += 1
        # hide appropriate sliders
        if self._selection == 0:  # C/W
            self.sl.set_cwvisible(True)
            self.setCropBounds()
        elif self._selection == 1:  # B/E
            self.sl.set_fcvisible(True)
            self.setCropBounds()
        elif self._selection == 2:  # slice
            self.sl.set_slicevisible(True)
            self.setSliceBounds()
        else:  # pass
            self.sl.set_allvisible(False)
            self.setPassBounds()


class ExternalNode(gpi.NodeAPI):
    """Display contents of Dictionary-type data

    INPUT
    Data of type DICT

    WIDGETS
    Keys: comma-delimited list of dictionary elements to display
    Range: allows inspection of a subset of large data arrays
    C/W - sliders control center/width of range
    B/E - sliders control begin/end of range
        Slice - slider gives element to inspect
        Pass - shows all elements
    """

    def execType(self):
        return gpi.GPI_THREAD

    def initUI(self):

        # Widgets
        self.addWidget('StringBox', 'Keys:')
        self.addWidget('PushButton', 'Match Exactly', val=0, toggle=True)
        self.addWidget('ReduceSliders', 'Range')
        self.addWidget('TextBox', 'Info:')

        self.addInPort('dq_in', 'DICT')

    def validate(self):
        '''update the slider bounds based on dictionary
        '''

        indat = self.getData('dq_in')
        inkeys = self.getVal('Keys:')
        exactmatch = self.getVal('Match Exactly')
        keys = gen_keys_from_keywords(inkeys, indat, exactmatch)

        w = self.getVal('Range')
        maxlen = 1
        if len(keys) > 0:
            dflat = dict()
            dflat = unwrap_dict(indat, dflat)
            for key in keys:
                if key in dflat:
                    try:
                        for subkey in list(dflat[key].keys()):
                            if('ndarray' in str(type(dflat[key][subkey])) or
                               'list' in str(type(dflat[key][subkey]))):
                                listlen = len(dflat[key][subkey])
                                if listlen > maxlen:
                                    maxlen = listlen
                    except:
                        if('ndarray' in str(type(dflat[key])) or
                           'list' in str(type(dflat[key]))):
                            listlen = len(dflat[key])
                            if listlen > maxlen:
                                maxlen = listlen
        else:
            for item in list(indat.items()):
                if('ndarray' in str(type(item[1]))
                   or 'list' in str(type(item[1]))):
                    listlen = len(item[1])
                    if listlen > maxlen:
                        maxlen = listlen
        self.setAttr('Range', max=maxlen)
        if w['selection'] == 3:  # pass
            w['center'] = (maxlen+1)/2
            w['width'] = maxlen
            w['floor'] = 1
            w['ceiling'] = maxlen
            self.setAttr('Range', quietval=w)

    def compute(self):

        import re

        indat = self.getData('dq_in')
        inkeys = self.getVal('Keys:')
        exactmatch = self.getVal('Match Exactly')
        keys = gen_keys_from_keywords(inkeys, indat, exactmatch)

        w = self.getVal('Range')
        minl = w['floor']-1
        maxl = w['ceiling']

        # Report Back to User
        report = '\n'
        if w['selection'] in [0, 1]:
            report = report + 'Displaying List Elements: ' + str(minl) + ':'\
                + str(maxl - 1) + '\n\n'
        elif w['selection'] == 2:
            report = report + 'Displaying Element: ' + str(minl) + '\n\n'
        if len(keys) > 0:
            dflat = dict()
            dflat = unwrap_dict(indat, dflat)
            for key in keys:
                if key in dflat:
                    if 'dict' in str(type(dflat[key])).lower():
                        report = report + str(key)+':\n'
                    try:
                        for item in list(sorted(dflat[key].items())):
                            report = print_item(item, report,
                                                minl, maxl)
                    except:
                        report = print_item((key, dflat[key]), report, minl,
                                            maxl)
        else:
            for item in list(sorted(indat.items())):
                report = print_item(item, report, minl, maxl)

        self.setAttr('Info:', val=report)

        return(0)




# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2013jul02

import numpy as np
import gpi

# define widget bounding limits
DIM_INT_MAX = pow(2, 31) - 1  # max QWidgets can use
DIM_INT_MIN = -pow(2, 31)  # min QWidgets can use

class ExternalNode(gpi.NodeAPI):
    """Peform operations on an array that generally change its shape and order, but not so much the actual data
    INPUT - input array
    OUTPUT - output array

    WIDGETS:
    Info: - information on size, etc of input and output array - very useful to pay attention to this
    Operation:
      Reshape - give new dimensions to file (must be same total size), while data stay in same order in memory
      Combine - combine 2 dimensions, output array will have one less dimension
      Split - split a dimension into two dimensions, output array will have one extra dimension
      Extend - add an extra dimension; if length of that dimension > 1, the data will be copied along that dimension
      Transpose - transpose any dimensions
      Flip - mirror the data along any dimension (can flip in multiple dimensions)
      Shift - shift data along any dimension, filling in with zeros
      CircShift - circularly shift data along any dimension, wrapping the data from the outgoing edge back around
      Tile - Separate a dimension so that it is tiled in the other dimensions - may not work in all cases (useful for 3D data)
    """

    def initUI(self):

        # Widgets
        self.dim_base_name = 'Dimension['
        self.maxdim = 13 # set to 13 for now
        self.ndim = self.maxdim
        self.shape = []
        self.size = 0
        self.comp = 1
        self.reset = 0
        self.trans_ind = []
        self.rep_ind = []
        self.flip_string = 'out = data[:]'
        self.info_message = ""
        self.op_buttons = ['Reshape', '   Combine', '   Split', 'Extend',
                           'Transpose', 'Flip', 'Shift', 'CircShift', 'Tile']
        self.addWidget('TextBox', 'Info:')
        self.addWidget('ExclusiveRadioButtons', 'Operation',
                       buttons=self.op_buttons, val = 0)
        self.addWidget('SpinBox', '# Dimensions', val=self.ndim)
        self.addWidget('SpinBox', '# Columns', val=1)
        self.addWidget('SpinBox', '# Rows', val=1)
        for i in range(self.maxdim):
            self.addWidget('SpinBox', self.dim_base_name+str(-i-1)+']', val=0)
        dim_buttons = []
        for i in range(-self.maxdim, 0):
            dim_buttons.append(str(i))
        self.addWidget('NonExclusivePushButtons', 'Selection',
                       buttons=dim_buttons, val=[])
        tile_buttons = []
        for i in range(-3, 0):
            tile_buttons.append(str(i))
        self.addWidget('ExclusivePushButtons', 'Tile Dimension',
                       buttons=tile_buttons, val=0)
                       
        self.addWidget('PushButton', 'Compute', toggle=True, val=0)
        self.addWidget('PushButton', 'Reset', toggle=True, val=0)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data = self.getData('in')
        op = self.getVal('Operation')
        basic_info = "Input Dimensions: "+str(data.shape)+"\n" \
                     "Input Size: "+str(data.size)+"\n"
        if data.ndim < 3:
            self.op_buttons = ['Reshape', '   Combine', '   Split', 'Extend',
                               'Transpose', 'Flip', 'Shift', 'CircShift']
            if op > 7:
                op = 0
            self.setAttr('Operation', buttons=self.op_buttons, val=op)
        else:
            self.op_buttons = ['Reshape', '   Combine', '   Split', 'Extend',
                               'Transpose', 'Flip', 'Shift', 'CircShift', 'Tile']
            self.setAttr('Operation', buttons=self.op_buttons, val=op)

        # generate list of any Dimension[?] widget events
        dim_events = [s for s in self.widgetEvents() if s.startswith(self.dim_base_name)]

        # set widget visibility and initial values
        if ('in' in self.portEvents() or 'Selection' in self.widgetEvents() or
            set(self.widgetEvents()).intersection(['Operation', 'Selection',
                'Reset', 'Tile Dimension', '# Rows', '# Columns'])):

            # initializations 
            self.flip_string = 'out = data[:]'
            self.info_message = ""

            # check for reset then turn it off
            self.reset = self.getVal('Reset')
            self.setAttr('Reset', val=0)

            # check to make sure the selected dimensions do not exceed the
            # data dimensionality
            if (self.op_buttons[op] in ['   Combine', '   Split',
                                        'Extend', 'Flip']):
                sel = self.getVal('Selection')
                if self.reset == 1:
                    sel = []
                if len(sel) > 0:
                    max_sel = max(sel)
                else:
                    max_sel = -1

            # get and setup data shape and dimensionality
            if op == 0 and self.reset == 0: # Reshape
                self.ndim = self.getVal('# Dimensions')
                self.shape = [1] * self.ndim
                for i in range(self.ndim):
                    self.shape[-i-1] = self.getVal(self.dim_base_name
                                                            +str(-i-1)+']')
                self.size = np.prod(self.shape)
                if self.size != data.size or self.size == 0:
                    self.ndim = data.ndim
                    self.shape = list(data.shape)

            elif op == 1: # Combine
                while max_sel >= len(data.shape)-1:
                    del sel[-1]
                    if len(sel) > 0:
                        max_sel = max(sel)
                    else:
                        break
                self.shape = list(data.shape)
                self.ndim = data.ndim
                for i in sel[::-1]:
                    for j in range(self.ndim - 1):
                        if j < i:
                            self.shape[j] = self.shape[j]
                        elif j == i:
                            self.shape[j] = (self.shape[j] * self.shape[j+1])
                        else:
                            self.shape[j] = self.shape[j+1]
                    del self.shape[-1]
                    self.ndim = len(self.shape)
                self.size = np.prod(self.shape)

            elif op == 2: # Split
                while max_sel >= len(data.shape):
                    del sel[-1]
                    if len(sel) > 0:
                        max_sel = max(sel)
                    else:
                        break
                # calculate indices of editable dimensions based on selection
                j = 0
                k = 0
                data_ind = list(range(len(data.shape) + len(sel)))
                self.shape = [1] * (len(data_ind))
                self.ndim = len(self.shape)
                vis_ind = [1] * (2 * len(sel))
                for i in sel:
                    data_ind.remove(i+k)
                    vis_ind[j] = i+k
                    vis_ind[j+1] = i+k+1
                    j = j+2
                    k = k+1
                # setup the data shape and dimensionality
                if self.portEvents():
                    for i in range(self.ndim):
                        self.shape[-i-1] = (self.getVal(
                                            self.dim_base_name+str(-i-1)+']'))
                    self.size = np.prod(self.shape)
                else:
                    self.reset = 1
                if self.size != data.size or self.reset == 1:
                    self.shape = [1] * self.ndim
                    k = 0
                    for i in range(self.ndim):
                        if i in data_ind:
                            self.shape[i] = data.shape[k]
                            k = k+1
                self.size = np.prod(self.shape)

            elif op == 3: # Extend
                while max_sel > len(data.shape):
                    del sel[-1]
                    if len(sel) > 0:
                        max_sel = max(sel)
                    else:
                        break
                # calculate indices of editable dimensions based on selection
                j = 0
                data_ind = list(range(len(data.shape) + len(sel)))
                self.shape = [1] * (len(data_ind))
                self.ndim = len(self.shape)
                vis_ind = [1] * (len(sel))
                for i in sel:
                    data_ind.remove(i+j)
                    vis_ind[j] = i+j
                    j = j+1
                self.rep_ind = vis_ind
                # setup the data shape and dimensionality
                if self.portEvents():
                    for i in range(self.ndim):
                        self.shape[-i-1] = (self.getVal(
                                            self.dim_base_name+str(-i-1)+']'))
                    self.size = np.prod(self.shape)
                j = 0
                for i in range(self.ndim):
                    if i in data_ind:
                        self.shape[i] = data.shape[j]
                        j = j+1
                self.size = np.prod(self.shape)

            elif self.op_buttons[op] == 'Tile': # Tile
                tile_dim = self.getVal('Tile Dimension')
                row_size = self.getVal('# Rows')
                column_size = self.getVal('# Columns')
                if '# Rows' in self.widgetEvents():
                    column_size = np.ceil(float(data.shape[tile_dim - 3]) / row_size)
                    self.setAttr('# Columns', val=column_size)
                if '# Columns' in self.widgetEvents():
                    row_size = np.ceil(float(data.shape[tile_dim - 3]) / column_size)
                    self.setAttr('# Rows', val=row_size)
                if (row_size * column_size < data.shape[tile_dim - 3] or 
                      self.reset == 1 or
                      'Tile Dimension' in self.widgetEvents()):
                    row_size = int(np.sqrt(data.shape[tile_dim - 3]))
                    column_size = np.ceil(float(data.shape[tile_dim - 3]) / row_size)
                    self.setAttr('# Rows', val=row_size)
                    self.setAttr('# Columns', val=column_size)

                self.ndim = len(data.shape) - 1
                self.shape = [1] * self.ndim
                if tile_dim == 0:
                    self.shape[-1] = data.shape[-1] * column_size
                    self.shape[-2] = data.shape[-2] * row_size
                elif tile_dim == 1:
                    self.shape[-1] = data.shape[-1] * column_size
                    self.shape[-2] = data.shape[-3] * row_size
                else:
                    self.shape[-1] = data.shape[-2] * column_size
                    self.shape[-2] = data.shape[-3] * row_size
                for i in range(3, data.ndim):
                    self.shape[-i] = data.shape[-i-1]

            elif self.op_buttons[op] == 'Transpose': # Transpose
                self.ndim = data.ndim
                self.shape = list(data.shape)
                self.trans_ind = list(range(-data.ndim, 0))
                if self.portEvents():
                    count_trans_ind = list(range(-data.ndim, 0))
                    for i in range(data.ndim):
                        trans_ind = self.getVal(self.dim_base_name
                                                         +str(-i-1)+']')
                        if trans_ind > data.ndim - 1 or trans_ind < -data.ndim:
                            trans_ind = -i-1
                        self.trans_ind[-i-1] = trans_ind
                        self.shape[-i-1] = data.shape[trans_ind]
                        if trans_ind < 0:
                            trans_ind = trans_ind + data.ndim
                        count_trans_ind[-i-1] = trans_ind
                    dup_count = max([count_trans_ind.count(i)
                                for i in count_trans_ind])
                    if dup_count > 1:
                        self.trans_ind = list(range(-data.ndim, 0))

            elif self.op_buttons[op] == 'Flip': # Flip
                self.shape = list(data.shape)
                self.ndim = data.ndim
                self.size = np.prod(data.shape)
                self.flip_string = 'out = data['
                while max_sel >= len(data.shape):
                    del sel[-1]
                    if len(sel) > 0:
                        max_sel = max(sel)
                    else:
                        break
                for i in range(data.ndim):
                    if i in sel:
                        self.flip_string = self.flip_string+'::-1, '
                    else:
                        self.flip_string = self.flip_string+':, '
                self.flip_string = self.flip_string[:-2]+']'

            else: # Shift or CircShift
                if 'Operation' in self.widgetEvents():
                    self.reset = 1
                self.shape = list(data.shape)
                self.ndim = data.ndim
                self.size = np.prod(data.shape)

            # change selection button labels based upon functionality
            dim_buttons = []
            if op == 1:
                for i in range(-data.ndim,-1):
                    dim_buttons.append(str(i)+" AND "+str(i+1))
            if self.op_buttons[op] in ['   Split', 'Flip']:
                for i in range(-data.ndim,0):
                    dim_buttons.append(str(i))
            if op == 3:
                for i in range(-data.ndim-1,0):
                    dim_buttons.append(str(i)+"->")

            # turn on selection button when used
            if (self.op_buttons[op] in ['   Combine', '   Split',
                                        'Extend', 'Flip']):
                self.setAttr('Selection', visible=True,
                                  buttons=dim_buttons, val=sel)
            else:
                self.setAttr('Selection', visible=False)

            # turn on tile widgets when used
            if self.op_buttons[op] == 'Tile':
                self.setAttr('# Rows', visible=True)
                self.setAttr('# Columns', visible=True)
                self.setAttr('Tile Dimension', visible=True)
            else:
                self.setAttr('# Rows', visible=False)
                self.setAttr('# Columns', visible=False)
                self.setAttr('Tile Dimension', visible=False)
                
            # dimensionality is only directly adjustable for reshape
            if op == 0:
                self.setAttr('# Dimensions', visible = True,
                                  val=self.ndim)
            else:
                self.setAttr('# Dimensions', visible = False,
                                  val=self.ndim)

            # set up Dimension widget values, visibility and range
            for i in range(self.maxdim):
                if i < self.ndim and op == 0: # Reshape
                    if self.reset == 1 or self.size == 0:
                        val = data.shape[-i-1]
                        if self.dim_base_name+str(-i-1)+']' not in dim_events:
                            self.setAttr(self.dim_base_name+str(-i-1)+']',
                                            visible=True, val=val,
                                            max=DIM_INT_MAX, min=1)
                        else:
                            self.setAttr(self.dim_base_name+str(-i-1)+']',
                                            visible=True,
                                            max=DIM_INT_MAX, min=1)
                    else:
                        val = self.getVal(self.dim_base_name+str(-i-1)+']')
                        self.shape[-i-1] = val
                        self.setAttr(self.dim_base_name+str(-i-1)+']',
                                          visible=True, val=val,
                                          max=DIM_INT_MAX, min=1)
                elif i < self.ndim and op == 1: # Combine
                    val = self.shape[-i-1]
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=False, val=val,
                                          max=DIM_INT_MAX, min=1)
                elif i < self.ndim and op in [2, 3]: # Split and Extend
                    val = self.shape[-i-1]
                    if (self.ndim - i - 1) in vis_ind:
                        if op == 2 and val == 1:
                            max_val = self.shape[-i]
                        elif op == 2:
                            max_val = val
                        else:
                            self.shape[-i-1] = 1
                            max_val = DIM_INT_MAX
                        self.setAttr(self.dim_base_name+str(-i-1)+']',
                                          visible=True, val=val, min=1,
                                          max=max_val)
                    else:
                        max_val = DIM_INT_MAX
                        self.setAttr(self.dim_base_name+str(-i-1)+']',
                                          visible=False, val=val, min=1,
                                          max=max_val)
                elif (i < data.ndim and
                      self.op_buttons[op] in ['Tile', 'Flip']):
                    val = data.shape[-i-1]
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=False, val=val, min=1, max=DIM_INT_MAX)
                elif i < data.ndim and self.op_buttons[op] == 'Transpose':
                    val = self.trans_ind[-i-1]
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=True, max=data.ndim-1,
                                      min=-data.ndim, val=val)
                elif (i < data.ndim and
                      self.op_buttons[op] in ['Shift', 'CircShift']):
                    if (self.reset == 1):
                        val = 0
                        self.setAttr(self.dim_base_name+str(-i-1)+']',
                                          visible=True, val=val,
                                          max=DIM_INT_MAX, min=DIM_INT_MIN)
                    else:
                        self.setAttr(self.dim_base_name+str(-i-1)+']',
                                          visible=True, max=DIM_INT_MAX,
                                          min=DIM_INT_MIN)
                else:
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=False, max=DIM_INT_MAX,
                                      min=DIM_INT_MIN)


        # adjust data and widget dimensionality when # dimensions is changed
        if ('# Dimensions' in self.widgetEvents() and op == 0):
            self.info_message = ""
            self.ndim = self.getVal('# Dimensions')
            if self.ndim >= len(self.shape):
                for i in range(len(self.shape), self.ndim):
                    self.setAttr(self.dim_base_name+str(-i-1)+']',
                                      visible=True, val=1)
                    self.shape.insert(0, 1)
            else:
                for i in range(len(self.shape), self.ndim, -1):
                    if self.shape[-i] == 1:
                        self.setAttr(self.dim_base_name+str(-i)+']',
                                          visible=False)
                        del self.shape[-i]
                    else:
                        self.ndim = len(self.shape)
                        self.setAttr('# Dimensions', val=self.ndim)
                        self.info_message = "Can't remove non-singleton "+ \
                                             "dimension\n"
                        break
            self.size = np.prod(self.shape)

        # adjust data output size based on user specified sizes
        if (dim_events):
            for event in dim_events:
                self.info_message = ""
                index = int(event[10:12])
                new_val = self.getVal(event)
                self.setAttr(event, val=new_val)
                if op == 0 and (self.ndim + index) >= 0:
                    self.shape[index] = new_val
                    self.size = np.prod(self.shape)
                    if self.size != data.size:
                        self.setAttr('Compute', val=0)
                        self.info_message = "The output and input sizes must "+ \
                                             "match\n"
                    elif self.comp == 1:
                        self.setAttr('Compute', val=1)
                if op == 2:
                    self.size = np.prod(self.shape)
                    sel = self.getVal('Selection')
                    k = 0
                    data_ind = list(range(self.ndim))
                    val = self.getVal(event)
                    for i in sel:
                        data_ind.remove(i+k)
                        k = k+1
                        old_val = self.shape[index]
                    if self.ndim + index in data_ind:
                        linked_index = index - 1
                    else:
                        linked_index = index + 1
                    dim_product = old_val * self.shape[linked_index]
                    val = new_val
                    while(np.mod(dim_product,val) != 0 and val > 1
                                 and val < dim_product):
                        if new_val > old_val:
                            val = val + 1
                        elif new_val < old_val:
                            val = val - 1
                    self.shape[index] = val
                    self.shape[linked_index] = dim_product / val
                    self.setAttr(event, val=val)
                    self.setAttr(self.dim_base_name+str(linked_index)+']',
                                      val=self.shape[linked_index])
                    self.size = np.prod(self.shape)
                if self.op_buttons[op] == 'Transpose':
                    count_trans_ind = list(range(-data.ndim, 0))
                    for i in range(data.ndim):
                        trans_ind = self.getVal(self.dim_base_name+
                                                         str(-i-1)+']')
                        self.trans_ind[-i-1] = trans_ind
                        self.shape[-i-1] = data.shape[trans_ind]
                        if trans_ind < 0:
                            trans_ind = trans_ind + data.ndim
                        count_trans_ind[-i-1] = trans_ind
                    dup_count = max([count_trans_ind.count(i)
                                    for i in count_trans_ind])
                    if dup_count > 1:
                        self.setAttr('Compute', val = 0)
                        self.info_message="Transpose indices must all be unique\n"
                    elif self.comp == 1:
                        self.setAttr('Compute', val=1)
                
        # disable compute when output and input data sizes are not equivalent
        if ('Compute' in self.widgetEvents()):
            self.info_message = ""
            if self.comp == 1:
                self.comp = 0
            else:
                self.comp = 1

        self.size = np.prod(self.shape)
        if self.size != data.size and op == 0:
            self.setAttr('Compute', val=0)
            self.info_message = "The output and input sizes must match\n"
        if self.op_buttons[op] == 'Transpose':
            count_trans_ind = list(range(-data.ndim, 0))
            for i in range(data.ndim):
                trans_ind = self.getVal(self.dim_base_name+
                                                 str(-i-1)+']')
                if trans_ind < 0:
                    trans_ind = trans_ind + data.ndim
                count_trans_ind[-i-1] = trans_ind
            dup_count = max([count_trans_ind.count(i)
                            for i in count_trans_ind])
            if dup_count > 1:
                self.setAttr('Compute', val = 0)
                self.info_message="Transpose indices must all be unique\n"

        self.reset = 0
        # updata info 
        info = basic_info+"Output Dimensions: "+str(self.shape)+"\n" \
               "Output Size: "+str(self.size)+"\n"+self.info_message
        self.setAttr('Info:', val = info)
        if( self.getVal('Compute')):
            self.setDetailLabel(self.op_buttons[op])
        else:
            self.setDetailLabel("{} (idle)".format(self.op_buttons[op]))

        return(0)

    def compute(self):

        import numpy as np

        data = self.getData('in')
        op = self.getVal('Operation')
        basic_info = "Input Dimensions: "+str(data.shape)+"\n" \
                     "Input Size: "+str(data.size)+"\n"

        if self.getVal('Compute'):
            if op in [0, 1, 2]: #Reshape, Combine, or Split
                out = data
                out.shape = self.shape
            if op == 3: # Extend
                temp = data
                temp.shape = self.shape
                for i in self.rep_ind:
                    reps = self.getVal(self.dim_base_name+
                                                str(i-len(self.shape))+']')
                    temp = np.repeat(temp, reps, axis = i)
                out = temp
            if self.op_buttons[op] == 'Tile': # Tile
                tile_dim = self.getVal('Tile Dimension')
                row_size = self.getVal('# Rows')
                column_size = self.getVal('# Columns')
                append_size = row_size * column_size - data.shape[tile_dim-3]
                append_shape = list(data.shape)
                append_shape[tile_dim-3] = append_size
                padded_data = np.append(data, np.zeros(append_shape),
                                        tile_dim-3)
                split_shape = [1] * (data.ndim + 1)
                tile_trans_ind = [1] * (data.ndim + 1)
                if tile_dim == 2:
                    tile_trans_ind[-4:] = [-2, -4, -1, -3]
                elif tile_dim == 1:
                    tile_trans_ind[-4:] = [-3, -4, -2, -1]
                else:
                    tile_trans_ind[-4:] = [-4, -2, -3, -1]

                for i in range(-data.ndim - 1, 0):
                    if i < tile_dim-4:
                        split_shape[i] = data.shape[i+1]
                    elif i == tile_dim-4:
                        split_shape[i] = row_size
                    elif i == tile_dim-3:
                        split_shape[i] = column_size
                    else:
                        split_shape[i] = data.shape[i]
                    if i < -4:
                        tile_trans_ind[i] = i
                padded_data.shape = split_shape
                out = np.ascontiguousarray(padded_data.transpose(
                                           tile_trans_ind))
                out.shape = self.shape
            if self.op_buttons[op] == 'Transpose': # Transpose
                out = data.transpose(self.trans_ind)
            if self.op_buttons[op] == 'Flip': # Flip
                g = globals()
                l = locals()
                exec(self.flip_string,g,l)
                out = l['out']
            if self.op_buttons[op] in ['Shift', 'CircShift']:
                temp = data
                for i in range(data.ndim):
                    shift = self.getVal(self.dim_base_name+
                                                 str(-i-1)+']')
                    if self.op_buttons[op] == 'Shift':
                        if shift <= 0:
                            shift = max(0, -shift)
                            del_range = list(range(shift))
                            insert_vals = [data.shape[-i-1]] * shift
                            temp = np.insert(temp, insert_vals, 0, axis = -i-1)
                            temp = np.delete(temp, del_range, axis = -i-1)
                        else:
                            shift = min(shift, data.shape[-i-1])
                            insert_vals = [0] * shift
                            del_range = list(range(data.shape[-i-1] - shift,
                                              data.shape[-i-1] + 1))
                            temp = np.delete(temp, del_range, axis = -i-1)
                            temp = np.insert(temp, insert_vals, 0, axis = -i-1)

                    if self.op_buttons[op] == 'CircShift':
                        temp = np.roll(temp, shift, axis = -i-1)
                out = temp
            self.shape = out.shape
            self.size = np.prod(self.shape)
            self.setData('out', out)

        # updata info 
        info = basic_info+"Output Dimensions: "+str(self.shape)+"\n" \
               "Output Size: "+str(self.size)+"\n"+self.info_message
        self.setAttr('Info:', val = info)

        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#Author: Jim Pipe
#Date: 2014Mar

import gpi
import numpy as np
from scipy import interpolate
from gpi import QtWidgets


class Interpolate_GROUP(gpi.GenericWidgetGroup):
    """A combination of SpinBoxes, DoubleSpinBoxes, and PushButtons
    to form a unique widget suitable for interpolation options on dimensions.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        # Widgets
        self._val = {}
        self._val['length'] = 1
        self._val['in_len'] = 1  # the original array length

        self.db = gpi.BasicDoubleSpinBox()  # factor
        self.db.set_label('factor:')
        self.db.set_min(0.001)
        self.db.set_max(gpi.GPI_FLOAT_MAX)
        self.db.set_decimals(8)
        self.db.set_singlestep(0.1)
        self.db.set_val(1)

        self.sb = gpi.BasicSpinBox()  # length
        self.sb.set_label('length:')
        self.sb.set_min(1)
        self.sb.set_val(1)
        self.sb.set_max(gpi.GPI_INT_MAX)

        self.db.valueChanged.connect(self.factChange)
        self.sb.valueChanged.connect(self.lenChange)

        vbox = QtWidgets.QHBoxLayout()
        vbox.addWidget(self.db)
        vbox.addWidget(self.sb)
        vbox.setStretch(0, 0)
        vbox.setStretch(1, 0)
        vbox.setStretch(2, 0)
        vbox.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        vbox.setSpacing(0)
        self.setLayout(vbox)

    def set_val(self, val):
        """A python-dict containing: in_len, length, and compute parms. """
        sig = False
        if 'in_len' in val:
            # otherwise this would change every time compute() was called
            if self._val['in_len'] != val['in_len']:
                self._val['in_len'] = val['in_len']
                fact = self.db.get_val()  # set len based on factor
                fact *= self._val['in_len']
                self.setLenQuietly(int(fact))
                self._val['length'] = int(fact)
        if 'length' in val:
            self._val['length'] = val['length']
            self.setLenQuietly(val['length'])
            self.setFactQuietly(float(val[
                                'length'])/float(self._val['in_len']))
            sig = True
        if 'compute' in val:
            self._val['compute'] = val['compute']
            sig = True
        if sig:
            self.valueChanged.emit()

    # getters
    def get_val(self):
        return self._val

    # support
    def factChange(self, val):
        self._val['length'] = int(self._val['in_len']*val)
        self.setLenQuietly(self._val['length'])
        self.valueChanged.emit()

    def lenChange(self, val):
        self._val['length'] = val
        self.setFactQuietly(float(val)/float(self._val['in_len']))
        self.valueChanged.emit()

    def compChange(self, val):
        self._val['compute'] = val
        self.valueChanged.emit()

    def setFactQuietly(self, val):
        self.db.blockSignals(True)
        self.db.set_val(val)
        self.db.blockSignals(False)

    def setLenQuietly(self, val):
        self.sb.blockSignals(True)
        self.sb.set_val(val)
        self.sb.blockSignals(False)

class ExternalNode(gpi.NodeAPI):
    """A node to linearly interpolate in specified directions.

    INPUT: input numpy array

    OUTPUT: interpolated data in the specified directions.

    WIDGETS:
    Dimension[n]:  For each dimension,
      factor - (desired output length)/(input length)
      length - desired output length
    kind - what type or order of interpolation to apply.
        'slinear', 'quadratic', and 'cubic' are all spline interpolations
        'sinc' interpolation uses zero-padding and FFTW
    Compute - compute
    """


    def initUI(self):

        # Widgets
        self.dim_base_name = 'Dimension['
        self.ndim = 6  # underlying c-code is only 6-dim
        for i in range(self.ndim):
            self.addWidget('Interpolate_GROUP', self.dim_base_name+str(i)+']')
        interp_modes =  ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'sinc')
        self.addWidget('ComboBox', 'interpolation-mode', items=interp_modes, val='slinear')
        windows =  ('none', 'hann', 'hamming', 'blackman', 'lanczos', 'tukey')
        self.addWidget('ComboBox', 'window', items=windows, val='lanczos', visible=False)
        self.addWidget('DoubleSpinBox', 'alpha', min=0.0, max=1.0, val=0.5, visible=False)
        self.addWidget('PushButton', 'compute', toggle=True)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addInPort('size', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        if 'in' in self.portEvents() or 'size' in self.portEvents():
            data = self.getData('in')
            out_shape = self.getData('size')

            # visibility and bounds
            for i in range(self.ndim):
                if i < len(data.shape):
                    val = {'in_len': data.shape[i]}

                    # assign output lengths based on optional input array
                    if out_shape is not None:
                        offset = len(data.shape)-len(out_shape.shape)
                        if offset < 0:
                            if i < len(out_shape.shape):
                                val['length'] = out_shape.shape[i-offset]
                        else:
                            if i >= offset:
                                val['length'] = out_shape.shape[i-offset]

                    self.setAttr(self.dim_base_name+str(i)+']', visible=True, quietval=val)
                else:
                    self.setAttr(self.dim_base_name+str(i)+']',visible=False)

        if self.getVal('interpolation-mode') == 'sinc':
            self.setAttr('window', visible=True)
            if self.getVal('window') == 'tukey':
                self.setAttr('alpha', visible=True)
            else:
                self.setAttr('alpha', visible=False)
        else:
            self.setAttr('window', visible=False)
            self.setAttr('alpha', visible=False)

        return(0)

    def compute(self):
        data_in = self.getData('in')
        kind = self.getVal('interpolation-mode')

        self.dimm = data_in.shape
        self.ndim = data_in.ndim

        data_out = data_in.copy()
        if self.getVal('compute'):
            if kind in ('linear', 'nearest'):
                for i in range(self.ndim):
                    val = self.getVal(self.dim_base_name+str(i)+']')
                    interpnew = val['length']
                    axisnew = self.dimm[i]
                    x = np.linspace(0, axisnew-1, axisnew)
                    xnew = np.linspace(0, axisnew-1, interpnew)
                    if axisnew == 1:
                        reps = np.ones((self.ndim,))
                        reps[i] = interpnew
                        ynew = np.tile(data_out, reps)
                    elif axisnew == interpnew:
                        continue
                    else:
                        yinterp = interpolate.interp1d(x, data_out,
                                                       kind=kind, axis=i)
                        ynew = yinterp(xnew)

                    data_out = ynew
            elif kind in ('zero', 'slinear', 'quadratic', 'cubic'):
                from scipy.ndimage.interpolation import map_coordinates
                orders = {'zero': 0, 'slinear': 1, 'quadratic': 2, 'cubic': 3}
                o = orders[kind]

                # if the data is complex, interp real and imaginary separately
                if data_in.dtype in (np.complex64, np.complex128):
                    data_real = np.real(data_in)
                    data_imag = np.imag(data_in)
                else:
                    data_real = data_in
                    data_imag = None

                new_dims = []
                for i in range(self.ndim):
                    val = self.getVal(self.dim_base_name + str(i) + ']')
                    new_dims.append(np.linspace(0, val['in_len']-1, val['length']))

                coords = np.meshgrid(*new_dims, indexing='ij')
                data_out = map_coordinates(data_real, coords, order=o)

                if data_imag is not None:
                    data_out = (data_out +
                                1j*map_coordinates(data_imag, coords, order=o))

            else:
                # use zero-padding and FFTW to sinc-interpolate
                import gpi_core.math.fft as ft
                data_in_c64 = np.require(data_in,
                                         dtype=np.complex64, requirements='C')

                old_dims = np.asarray(self.dimm[::-1], dtype=np.int64)
                new_dims = old_dims.copy()
                fftargs = {}
                win = np.ones(data_in.shape)
                scale = 1
                for i in range(self.ndim):
                    val = self.getVal(self.dim_base_name + str(i) + ']')
                    new_dims[self.ndim-i-1] = np.int64(val['length'])
                    if val['length'] == val['in_len']:
                        fftargs['dim{}'.format(self.ndim-i)] = 0
                    else:
                        win *= self.window(i)
                        fftargs['dim{}'.format(self.ndim-i)] = 1
                        scale *= val['length'] / val['in_len']

                # forward FFT with original dimensions
                fftargs['dir'] = 0
                data_out = ft.fftw(data_in_c64, old_dims, **fftargs)

                data_out *= win.astype(np.complex64)
                # inverse FFT with new dimensions (zero-pad kspace)
                fftargs['dir'] = 1
                data_out = ft.fftw(data_out, new_dims, **fftargs)

                if data_in.dtype in (np.float32, np.float64):
                    data_out = np.real(data_out)

                # data needs to be scaled since we changed the size between the
                # forward and inverse FT
                data_out *= scale

            self.setData('out', data_out)
        else:
            pass

        return(0)


    def execType(self):
        return gpi.GPI_PROCESS

    def window(self, axis):
        kind = self.getVal('window')
        N = self.dimm[axis]

        win = np.ones(N) # kind = 'none' is a rectangular window

        if kind == 'hann':
            win = np.hanning(N)
        elif kind == 'hamming':
            win = np.hamming(N)
        elif kind == 'blackman':
            win = np.blackman(N)
        elif kind == 'lanczos':
            n = np.arange(N)
            win = np.sinc(2. * np.arange(N) / (N - 1) - 1)
        elif kind == 'tukey':
            a = self.getVal('alpha')
            if a != 0.0:
                l_edge = a * (N - 1) // 2
                r_edge = round((N - 1) * (1 - a / 2))
                n = np.arange(N)
                win[0:l_edge] = 0.5*(1 +
                                np.cos(np.pi * (2*n[0:l_edge] / a / (N - 1) - 1)))
                win[l_edge:r_edge] = 1
                win[r_edge::] = 0.5*(1 +
                                np.cos(np.pi * (2*n[r_edge::] / a / (N - 1)  - 2/a + 1)))

        for ax in range(self.ndim):
            if ax < axis:
                win = win[np.newaxis, ...]
            elif ax > axis:
                win = win[..., np.newaxis]

        return win



